import unittest
import torch
from irisml.tasks.get_topk import Task


class TestGetTopk(unittest.TestCase):
    def test_2d(self):
        tensor = torch.randn(10, 24)
        outputs = Task(Task.Config(3, device='cpu')).execute(Task.Inputs(tensor))
        self.assertEqual(outputs.values.shape, (10, 3))
        self.assertEqual(outputs.indices.shape, (10, 3))

        topk_results = torch.topk(tensor, 3)
        self.assertTrue(torch.all(outputs.values == topk_results.values))
        self.assertTrue(torch.all(outputs.indices == topk_results.indices))

    def test_3d(self):
        tensor = torch.randn(10, 8, 24)
        outputs = Task(Task.Config(3, device='cpu')).execute(Task.Inputs(tensor))
        self.assertEqual(outputs.values.shape, (10, 8, 3))
        self.assertEqual(outputs.indices.shape, (10, 8, 3))

        topk_results = torch.topk(tensor, 3)
        self.assertTrue(torch.all(outputs.values == topk_results.values))
        self.assertTrue(torch.all(outputs.indices == topk_results.indices))

    def test_dry_run(self):
        tensor = torch.randn(10, 24)
        outputs = Task(Task.Config(3, device='cpu')).dry_run(Task.Inputs(tensor))
        self.assertEqual(outputs.values.shape, (10, 3))
        self.assertEqual(outputs.indices.shape, (10, 3))



import dataclasses
import logging
import irisml.core
from typing import Union, List, Optional

logger = logging.getLogger(__name__)


class Task(irisml.core.TaskBase):
    """ pick from vals based on conditions. Task will return the first val with condition being True.
    default will be returned, if all conditions are false.

    Inputs:
        conditions (bool): conditions
        vals (float, int, str): vals
        default (float, int, str): default value
    """
    VERSION = '0.1.0'

    @dataclasses.dataclass
    class Inputs:
        conditions: List[bool]
        vals: List[Union[float, int, str, list]]
        default: Optional[Union[float, int, str, list]] = None

    @dataclasses.dataclass
    class Outputs:
        result: Union[float, int, str, list]

    def execute(self, inputs):
        assert len(inputs.conditions) == len(inputs.vals)

        for i, cond in enumerate(inputs.conditions):
            if cond:
                logger.info(f'{i}th condition is True and got picked.')
                return self.Outputs(inputs.vals[i])

        logger.info(f'None of conditions is True: default {inputs.default} returned.')
        return self.Outputs(inputs.default)

    def dry_run(self, inputs):
        return self.execute(inputs)



import contextlib
import io
import sys
import unittest
import unittest.mock
import irisml.core
from irisml.tasks.branch import Task


class TestBranch(unittest.TestCase):
    def test_single_task(self):
        class FakeTask(irisml.core.TaskBase):
            def execute(self, inputs):
                print("fake_task is called.")
                return self.Outputs()

        class FakeTask2(irisml.core.TaskBase):
            def execute(self, inputs):
                print("fake_task2 is called.")
                return self.Outputs()

        with unittest.mock.patch.dict(sys.modules, {'irisml.tasks.fake_task': unittest.mock.MagicMock(Task=FakeTask),
                                                    'irisml.tasks.fake_task2': unittest.mock.MagicMock(Task=FakeTask2)}):
            with contextlib.redirect_stdout(io.StringIO()) as f:
                Task(Task.Config(True, [irisml.core.TaskDescription.from_dict({'task': 'fake_task'})])).execute(Task.Inputs())
            self.assertEqual(f.getvalue().strip(), 'fake_task is called.')

            with contextlib.redirect_stdout(io.StringIO()) as f:
                Task(Task.Config(True,
                                 [irisml.core.TaskDescription.from_dict({'task': 'fake_task'})],
                                 [irisml.core.TaskDescription.from_dict({'task': 'fake_task2'})])).execute(Task.Inputs())
            self.assertEqual(f.getvalue().strip(), 'fake_task is called.')

            with contextlib.redirect_stdout(io.StringIO()) as f:
                Task(Task.Config(False,
                                 [irisml.core.TaskDescription.from_dict({'task': 'fake_task'})],
                                 [irisml.core.TaskDescription.from_dict({'task': 'fake_task2'})])).execute(Task.Inputs())
            self.assertEqual(f.getvalue().strip(), 'fake_task2 is called.')

    def test_dry_run(self):
        class FakeTask(irisml.core.TaskBase):
            def execute(self, inputs):
                raise RuntimeError()  # This method must not be called in dry_run.

        with unittest.mock.patch.dict(sys.modules, {'irisml.tasks.fake_task': unittest.mock.MagicMock(Task=FakeTask)}):
            outputs = Task(Task.Config(True, [irisml.core.TaskDescription.from_dict({'task': 'fake_task'})])).dry_run(Task.Inputs())
            self.assertIsNotNone(outputs)



import dataclasses
import logging
import multiprocessing
import typing
import queue
import irisml.core

logger = logging.getLogger(__name__)


class Task(irisml.core.TaskBase):
    """Run the given tasks in parallel. A new process will be forked for each task. Each task must have an unique name."""
    VERSION = '0.1.0'
    CACHE_ENABLED = False

    @dataclasses.dataclass
    class Config:
        tasks: typing.List[irisml.core.TaskDescription]

    def execute(self, inputs):
        tasks = [irisml.core.Task(t) for t in self.config.tasks]
        names_set = set(t.name for t in tasks)

        if len(names_set) != len(tasks):
            raise ValueError("The child task names must be unique.")

        def run(task, context, q, index):
            task.load_module()
            outputs = task.execute(context)
            q.put((index, outputs))

        mp_ctx = multiprocessing.get_context('fork')
        q = mp_ctx.Queue(len(tasks))
        processes = [mp_ctx.Process(target=run, args=(t, self.context, q, i)) for i, t in enumerate(tasks)]
        logger.debug(f"Created {len(processes)} processes.")

        for p in processes:
            p.start()

        for p in processes:
            p.join()

        for i in range(len(tasks)):
            try:
                index, outputs = q.get_nowait()
            except queue.Empty:
                raise RuntimeError("Failed to retrieve child outputs.")
            self.context.add_outputs(tasks[index].name, outputs)

        return self.Outputs()

    def dry_run(self, inputs):
        return self.execute(inputs)



import dataclasses
import logging
import random
import typing
import irisml.core
import torch
import PIL.Image

logger = logging.getLogger(__name__)


class Task(irisml.core.TaskBase):
    """Generate a fake image classification dataset.

    A generated dataset returns (image:PIL.Image, target:int)

    Config:
        num_images (int): Number of images to generate. Default: 100.
        num_classes (int): Number of classes to generate. Default: 10.
        random_seed (int): Random seed. Default: 0.
    """
    VERSION = '0.1.2'

    @dataclasses.dataclass
    class Config:
        num_images: int = 100
        num_classes: int = 10
        random_seed: int = 0

    @dataclasses.dataclass
    class Outputs:
        dataset: torch.utils.data.Dataset
        num_classes: int
        class_names: typing.List[str]

    class FakeDataset(torch.utils.data.Dataset):
        def __init__(self, image_labels: typing.List[int], colors: typing.List[typing.Tuple[int, int, int]]):
            self._image_labels = image_labels
            self._colors = colors

        def __len__(self):
            return len(self._image_labels)

        def __getitem__(self, index):
            image_class = self._image_labels[index]
            return self._generate_image(image_class), torch.tensor(image_class)

        def _generate_image(self, class_id):
            return PIL.Image.new('RGB', (224, 224), color=self._colors[class_id])

    def execute(self, inputs):
        if not self.config.num_images or not self.config.num_classes:
            raise ValueError(f"Invalid settings: num_images={self.config.num_images}, num_classes={self.config.num_classes}")

        if self.config.num_images < self.config.num_classes:
            logger.info(f"Some classes will be empty: num_images={self.config.num_images}, num_classes={self.config.num_classes}")

        num_images_per_class = [self.config.num_images // self.config.num_classes for _ in range(self.config.num_classes)]
        for i in range(self.config.num_images % self.config.num_classes):
            num_images_per_class[i] += 1

        assert sum(num_images_per_class) == self.config.num_images
        assert len(num_images_per_class) == self.config.num_classes

        labels = [c for c, x in enumerate(num_images_per_class) for _ in range(x)]
        rng = random.Random(self.config.random_seed)
        rng.shuffle(labels)

        colors = [(rng.randint(0, 255), rng.randint(0, 255), rng.randint(0, 255)) for _ in range(self.config.num_classes)]
        logger.debug(f"Generated labels: {colors}")
        dataset = Task.FakeDataset(labels, colors)
        class_names = [f'class_{i}' for i in range(self.config.num_classes)]
        return self.Outputs(dataset, num_classes=self.config.num_classes, class_names=class_names)

    def dry_run(self, inputs):
        return self.execute(inputs)



import unittest
import pathlib
import tempfile
import PIL.Image
import torch
from irisml.tasks.save_images_from_dataset import Task


class TestSaveImagesFromDataset(unittest.TestCase):
    def test_no_indices(self):
        dataset = [(PIL.Image.new('RGB', (10, 10)), torch.tensor(0)), (PIL.Image.new('RGB', (10, 10)), torch.tensor(1))]
        with tempfile.TemporaryDirectory() as temp_dirpath:
            temp_dirpath = pathlib.Path(temp_dirpath) / 'images'  # Nonexistent directory
            Task(Task.Config(dirpath=temp_dirpath)).execute(Task.Inputs(dataset=dataset))

            self.assertTrue((temp_dirpath / '0.png').exists())
            self.assertTrue((temp_dirpath / '1.png').exists())
            self.assertFalse((temp_dirpath / '2.png').exists())

            image0 = PIL.Image.open(temp_dirpath / '0.png')
            self.assertEqual(image0.size, (10, 10))

    def test_save_with_indices(self):
        dataset = [(PIL.Image.new('RGB', (10, 10)), torch.tensor(0)), (PIL.Image.new('RGB', (10, 10)), torch.tensor(1))] * 2
        with tempfile.TemporaryDirectory() as temp_dirpath:
            temp_dirpath = pathlib.Path(temp_dirpath) / 'images'  # Nonexistent directory
            Task(Task.Config(dirpath=temp_dirpath, indices=torch.tensor([0, 3]))).execute(Task.Inputs(dataset=dataset))

            self.assertTrue((temp_dirpath / '0.png').exists())
            self.assertFalse((temp_dirpath / '1.png').exists())
            self.assertFalse((temp_dirpath / '2.png').exists())
            self.assertTrue((temp_dirpath / '3.png').exists())

            image0 = PIL.Image.open(temp_dirpath / '0.png')
            self.assertEqual(image0.size, (10, 10))



import dataclasses
import itertools
import logging
import typing
import irisml.core


logger = logging.getLogger(__name__)


@dataclasses.dataclass
class SearchSpaceConfig:
    name: str
    candidates: typing.List[typing.Union[int, float, str]]


class Task(irisml.core.TaskBase):
    """Grid search hyperparameters. Tasks are run in sequence.

       In the search loop, new combination of hyperparameters will be set to environment variables.

       After the search loop, the best iteration will be selected and added to the context.
    """
    VERSION = '0.1.0'
    CACHE_ENABLED = False
    RESOLVE_CONFIG_VARIABLES = False

    @dataclasses.dataclass
    class Config:
        search_space: typing.List[SearchSpaceConfig]
        tasks: typing.List[irisml.core.TaskDescription]
        metrics_output_name: str

    @dataclasses.dataclass
    class Outputs:
        best_parameters: typing.Dict[str, typing.Union[int, float, str]] = dataclasses.field(default_factory=dict)

    def execute(self, inputs):
        return self.execute_task(inputs, False)

    def execute_task(self, inputs, dry_run):
        if not self.config.search_space:
            return self.Outputs()

        parameter_names = [x.name for x in self.config.search_space]
        parameter_combinations = list(itertools.product(*[x.candidates for x in self.config.search_space]))

        logger.info(f"Trying {len(parameter_combinations)} combinations...")

        tasks = []
        for task_description in self.config.tasks:
            task = irisml.core.Task(task_description)
            task.load_module()
            tasks.append(task)

        best_results = None  # Metrics, Context object
        metrics_output_task_name, metrics_output_variable_name = self.config.metrics_output_name.split('.')
        for params in parameter_combinations:
            selected_parameters = dict(zip(parameter_names, params))
            logger.info(f"Trying params: {selected_parameters}")
            context = self.context.clone()
            for name, value in selected_parameters.items():
                context.add_environment_variable(name, value)

            for task in tasks:
                task.execute(context) if not dry_run else task.dry_run(context)

            metrics = getattr(context.get_outputs(metrics_output_task_name), metrics_output_variable_name)
            logger.debug(f"Result was {metrics} with {selected_parameters}")
            if not best_results or best_results[0] < metrics:
                best_results = (metrics, selected_parameters, context)

        logger.info(f"Best result is {best_results[0]} with parameters {best_results[1]}")

        # Add the outputs from the best trial to the current context.
        for task in tasks:
            self.context.add_outputs(task.name, best_results[2].get_outputs(task.name))

        return self.Outputs(best_results[1])

    def dry_run(self, inputs):
        return self.execute_task(inputs, True)



import dataclasses
import logging
import irisml.core
from typing import Union, Optional

logger = logging.getLogger(__name__)


class Task(irisml.core.TaskBase):
    """ Compare two values
    when is_equal=true, it tests for equality, otherwise it tests for greater than or less than.

    Config:
        is_equal (bool): Whether the two values should be equal.
        greater (bool): Whether the first value should be greater than the second.
        equal_allowed (bool): Whether equal values are allowed, when testing for greater than or less than.
    Inputs:
        val1 (float, int, str): value 1
        val2 (float, int, str): value 2
    """
    VERSION = '0.1.1'

    @dataclasses.dataclass
    class Inputs:
        val1: Union[float, int, str]
        val2: Union[float, int, str]

    @dataclasses.dataclass
    class Outputs:
        result: bool

    @dataclasses.dataclass
    class Config:
        equal_allowed: Optional[bool] = True
        greater: Optional[bool] = True
        is_equal: Optional[bool] = False

    def execute(self, inputs):
        if self.config.is_equal:
            return self.Outputs(inputs.val1 == inputs.val2)
        if self.config.equal_allowed and inputs.val1 == inputs.val2:
            return self.Outputs(True)

        return self.Outputs(inputs.val1 > inputs.val2 if self.config.greater else inputs.val1 < inputs.val2)

    def dry_run(self, inputs):
        return self.execute(inputs)



import dataclasses
import logging
import random
import typing
import irisml.core
import torch
import PIL.Image
import PIL.ImageColor
import PIL.ImageDraw

logger = logging.getLogger(__name__)


class Task(irisml.core.TaskBase):
    """Generate a fake phrase grounding dataset.

    An image has a random number of objects. Each object is a bounding box with a random color.

    Configs:
        num_images (int): The number of images to make. Default: 100
        num_max_boxes (int): The max number of objects per image. Default: 10
        image_sizes ([(int, int), ...]): A list of candidate image sizes. Default: [(320, 320)]
        random_seed (int): The random seed. Default: 0
    """
    VERSION = '0.1.0'

    @dataclasses.dataclass
    class Config:
        num_images: int = 100
        num_max_boxes: int = 10  # The max number of objects per image.
        image_sizes: typing.List[typing.Tuple[int, int]] = dataclasses.field(default_factory=lambda: [(320, 320)])
        random_seed: int = 0

    @dataclasses.dataclass
    class Outputs:
        dataset: torch.utils.data.Dataset

    class FakeDataset(torch.utils.data.Dataset):
        def __init__(self, targets: typing.List[typing.Tuple[str, typing.List[typing.List]]], sizes: typing.List[typing.Tuple[int, int]]):
            assert len(targets) == len(sizes)
            self._targets = targets
            self._sizes = sizes

        def __len__(self):
            return len(self._targets)

        def __getitem__(self, index):
            targets = self._targets[index]
            size = self._sizes[index]

            image = PIL.Image.new('RGB', size)
            draw = PIL.ImageDraw.Draw(image)
            for t in targets:
                draw.rectangle((t[1][0] * size[0], t[1][1] * size[1], t[1][2] * size[0], t[1][3] * size[1]), fill=t[0])
            box_text = [f'a {t[0]} box' for t in targets]
            caption = 'This image has ' + ', '.join(box_text)
            current = len('This image has ')
            text_span = []
            for text in box_text:
                text_span.append((current, current + len(text)))
                assert caption[current:current + len(text)] == text
                current += len(text) + 2

            new_targets = [(span, torch.tensor([t[1]])) for span, t in zip(text_span, targets)]
            return (caption, image), new_targets

    def execute(self, inputs):
        rng = random.Random(self.config.random_seed)

        def _random_box(image_width, image_height):
            # Make sure the bounding box size is at least 1 pixel.
            cx = rng.uniform(1 / image_width / 2, 1 - 1 / image_width / 2)
            cy = rng.uniform(1 / image_height / 2, 1 - 1 / image_height / 2)
            w = rng.uniform(1 / image_width, min(1 - cx, cx) * 2)
            h = rng.uniform(1 / image_height, min(1 - cy, cy) * 2)
            return [cx - w / 2, cy - h / 2, cx + w / 2, cy + h / 2]

        if not self.config.image_sizes:
            raise ValueError("Image sizes setting is empty.")

        for c in self.config.image_sizes:
            if len(c) != 2 or c[0] <= 0 or c[1] <= 0:
                raise ValueError(f"Invalid image size configs: {self.config.image_sizes}")

        sizes = rng.choices(self.config.image_sizes, k=self.config.num_images)
        color_names = list(PIL.ImageColor.colormap.keys())
        targets = []
        for i in range(self.config.num_images):
            boxes = [_random_box(*sizes[i]) for _ in range(rng.randrange(self.config.num_max_boxes + 1))]
            colors = rng.choices(color_names, k=len(boxes))
            targets.append([(color, box) for color, box in zip(colors, boxes)])
        dataset = Task.FakeDataset(targets, sizes)
        return self.Outputs(dataset)

    def dry_run(self, inputs):
        return self.execute(inputs)



import collections
import dataclasses
import json
import logging
import re
import typing
import torch
import irisml.core

logger = logging.getLogger(__name__)


class Task(irisml.core.TaskBase):
    """Get an integer from a JSON string.

    This task takes a list of JSON strings and returns a list of integers. The
    JSON strings are expected to be dictionaries with a key that matches the
    `key_name` config value. If the JSON string cannot be decoded or the key
    cannot be found, the value -1 is returned.

    This task ignores any non-JSON text before or after the JSON string.

    If single quotes are used in the JSON string, they are replaced with double
    quotes before decoding.

    Config:
        key_name (str): The key to look for in the JSON string.
        voting (bool): If True, the most common value is returned. If False, the first value is returned.
    """
    VERSION = '1.1.2'

    @dataclasses.dataclass
    class Inputs:
        json_strings: typing.List[str]

    @dataclasses.dataclass
    class Config:
        key_name: str
        voting: bool = False

    @dataclasses.dataclass
    class Outputs:
        tensor: torch.Tensor

    def execute(self, inputs):
        ints = []
        for i, json_string in enumerate(inputs.json_strings):
            value = -1
            try:
                values = self._get_ints_from_json(json_string)
                if self.config.voting:
                    c = collections.Counter(values)
                    value = collections.Counter(values).most_common(1)[0][0]
                    if len(c) > 1:
                        logger.debug(f"Index {i}: Picked {value} from {values}")
                else:
                    value = values[0]

            except Exception as e:
                logger.warning(f"Index {i}: {repr(e)}. json: {repr(json_string)}")

            logger.info(f"Index {i}: Found value {value} in json: {repr(json_string)}")
            ints.append(value)

        int_tensor = torch.tensor(ints)
        logger.info(f"Returning {int_tensor.shape} tensor of ints")
        return self.Outputs(int_tensor)

    def dry_run(self, inputs):
        return self.execute(inputs)

    def _get_ints_from_json(self, json_str):
        index = 0
        values = []

        while index < len(json_str):
            json_dict, end_index = self._decode_json(json_str[index:])
            if json_dict is None:
                break
            assert end_index > 0
            index += end_index

            if self.config.key_name not in json_dict:
                logger.warning(f"Key {self.config.key_name} not found in JSON string: {repr(json_str[index:])}")
                continue

            if not isinstance(json_dict[self.config.key_name], int):
                logger.warning(f"Value {json_dict[self.config.key_name]} is not an integer")
                continue

            values.append(json_dict[self.config.key_name])

        if not values:
            logger.warning("All attempts to decode JSON string failed. Trying to extract values without decoding.")
            values = self._try_extract_values_directly(json_str, self.config.key_name)
            if not values:
                raise ValueError("No values found in JSON string.")

        return values

    @staticmethod
    def _decode_json(json_str):
        """Decode a JSON string.

        This method tries to decode a JSON string. If the string cannot be
        decoded, it tries to fix the string and decode it again. If the string
        still cannot be decoded, it returns None.

        Returns:
            A tuple of the decoded JSON object and the index of the end of the
            JSON string in the original string.
        """
        start = json_str.find('{')
        if start == -1:
            return None, len(json_str)

        try:
            return json.JSONDecoder().raw_decode(json_str[start:])
        except json.JSONDecodeError:
            pass

        try:
            # Try replacing a single quote with a double quote and vice versa.
            trans = str.maketrans('\'"', '"\'')
            return json.JSONDecoder().raw_decode(json_str.translate(trans)[start:])
        except json.JSONDecodeError:
            pass

        try:
            # Try replacing a single quote with a double quote
            return json.JSONDecoder().raw_decode(json_str.replace('\'', '"')[start:])
        except json.JSONDecodeError:
            pass

        try:
            # Try escaping a double quote in the JSON string
            pattern = r'([:\[,{]\s*)"(.*?)"(?=\s*[:,\]}])'

            def replace(match):
                return match.group(1) + '"' + match.group(2).replace('"', "'") + '"'
            fixed_json = re.sub(pattern, replace, json_str[start:])
            return json.JSONDecoder().raw_decode(fixed_json)
        except json.JSONDecodeError:
            return None, len(json_str)

        return None, len(json_str)

    @staticmethod
    def _try_extract_values_directly(json_str, key_name):
        matches = re.findall(rf'"{key_name}"\s*:\s*(\-?\d+)', json_str)
        return [int(m) for m in matches]



import unittest
import torch
from irisml.tasks.calculate_cosine_similarity import Task


class TestCalculcateCosineSimilarity(unittest.TestCase):
    def test_identical_vectors(self):
        one_vector = torch.randn(1, 64)
        tensor0 = one_vector.expand(10, 64)
        tensor1 = one_vector.expand(8, 64)
        outputs = Task(Task.Config(device='cpu')).execute(Task.Inputs(tensor0, tensor1))
        self.assertEqual(outputs.cosine_similarity.shape, (10, 8))
        max_diff = torch.max(torch.abs(outputs.cosine_similarity - 1.0))
        self.assertAlmostEqual(float(max_diff), 0, places=6)  # Assert that cosine_similarity is 1 for all combinations.

    def test_zeros(self):
        tensor = torch.zeros((10, 64))
        outputs = Task(Task.Config(device='cpu')).execute(Task.Inputs(tensor, tensor))
        self.assertEqual(outputs.cosine_similarity.shape, (10, 10))
        self.assertAlmostEqual(float(torch.max(torch.abs(outputs.cosine_similarity))), 0)

    def test_random(self):
        tensor0 = torch.randn((3, 64))
        tensor1 = torch.randn((5, 64))
        outputs = Task(Task.Config(device='cpu')).execute(Task.Inputs(tensor0, tensor1))
        self.assertTrue(torch.all((-1 <= outputs.cosine_similarity) & (outputs.cosine_similarity <= 1)))



import dataclasses
import random
import typing
import irisml.core


class Task(irisml.core.TaskBase):
    """Make a text transform function that randomly chooses one of the substrings separated by the delimiter.

    Config:
        delimiter (str): The delimiter to split the text by.
    """
    VERSION = '0.1.0'

    @dataclasses.dataclass
    class Config:
        delimiter: str = '<|>'

    @dataclasses.dataclass
    class Outputs:
        transform: typing.Callable[[str], str]

    def execute(self, inputs):
        return self.Outputs(transform=RandomChoiceTransform(self.config.delimiter))

    def dry_run(self, inputs):
        return self.execute(inputs)


class RandomChoiceTransform:
    def __init__(self, delimiter):
        self._delimiter = delimiter

    def __call__(self, text):
        return random.choice(text.split(self._delimiter))



import unittest
from irisml.tasks.make_text_transform import Task


class TestMakeTextTransform(unittest.TestCase):
    def test_simple(self):
        outputs = Task(Task.Config('Hello, {}!')).execute(Task.Inputs())
        self.assertEqual(outputs.transform('world'), 'Hello, world!')

    def test_invalid_template(self):
        with self.assertRaises(ValueError):
            Task(Task.Config('Hello, world!')).execute(Task.Inputs())



import dataclasses
import sys
import unittest
import unittest.mock
import irisml.core
from irisml.tasks.repeat_tasks import Task


class FakeTask:
    class Task(irisml.core.TaskBase):
        @dataclasses.dataclass
        class Inputs:
            int_value: int

        @dataclasses.dataclass
        class Outputs:
            float_value: float

        def execute(self, inputs):
            return self.Outputs(float(inputs.int_value * 2))


class TestRepeatTasks(unittest.TestCase):
    def test_no_outputs(self):
        with unittest.mock.patch.dict(sys.modules, {'irisml.tasks.fake_task': FakeTask}):
            config = Task.Config(tasks=[irisml.core.TaskDescription.from_dict({'task': 'fake_task', 'inputs': {'int_value': '$env.REPEAT_TASKS_INDEX'}})],
                                 num_repeats=3)
            outputs = Task(config).execute(Task.Inputs())
            self.assertEqual(outputs.float_output_values, [[], [], []])

    def test_collect_outputs(self):
        with unittest.mock.patch.dict(sys.modules, {'irisml.tasks.fake_task': FakeTask}):
            config = Task.Config(tasks=[irisml.core.TaskDescription.from_dict({'task': 'fake_task', 'inputs': {'int_value': '$env.REPEAT_TASKS_INDEX'}})],
                                 num_repeats=3,
                                 float_output_names=['fake_task.float_value'])
            outputs = Task(config).execute(Task.Inputs())
            self.assertEqual(outputs.float_output_values, [[0.0], [2.0], [4.0]])



import unittest
import torch.nn
from irisml.tasks.check_model_parameters import Task


class TestCheckModelParameters(unittest.TestCase):
    def test_ok_model(self):
        model = torch.nn.Conv2d(3, 3, 3)
        outputs = Task(Task.Config(throw_exception=False)).execute(Task.Inputs(model))
        self.assertFalse(outputs.has_nan_or_inf)

    def test_nan_model(self):
        model = torch.nn.Conv2d(3, 3, 3)
        model.weight.data = model.weight.data * float('nan')
        outputs = Task(Task.Config(throw_exception=False)).execute(Task.Inputs(model))
        self.assertTrue(outputs.has_nan_or_inf)

        model.weight.data = model.weight.data * float('inf')
        outputs = Task(Task.Config(throw_exception=False)).execute(Task.Inputs(model))
        self.assertTrue(outputs.has_nan_or_inf)

    def test_throw_exception(self):
        model = torch.nn.Conv2d(3, 3, 3)
        model.weight.data = model.weight.data * float('nan')
        with self.assertRaises(ValueError):
            Task(Task.Config(throw_exception=True)).execute(Task.Inputs(model))



import pathlib
import tempfile
import unittest
from irisml.tasks.load_strs_from_text_file import Task


class TestLoadStrsFromTextFile(unittest.TestCase):
    def test_simple(self):
        with tempfile.NamedTemporaryFile() as f:
            temp_filepath = pathlib.Path(f.name)

            temp_filepath.write_text('hello\nworld')
            outputs = Task(Task.Config(filepath=temp_filepath)).execute(Task.Inputs())
            self.assertEqual(outputs.strs, ['hello', 'world'])
            self.assertEqual(outputs.num_strs, 2)

            # With a trailing newline
            temp_filepath.write_text('hello\nworld \n')
            outputs = Task(Task.Config(filepath=temp_filepath)).execute(Task.Inputs())
            self.assertEqual(outputs.strs, ['hello', 'world'])
            self.assertEqual(outputs.num_strs, 2)

            # With a trailing newline and empty line
            temp_filepath.write_text('hello\nworld\n\n')
            outputs = Task(Task.Config(filepath=temp_filepath)).execute(Task.Inputs())
            self.assertEqual(outputs.strs, ['hello', 'world'])
            self.assertEqual(outputs.num_strs, 2)

            # empty file
            temp_filepath.write_text('')
            outputs = Task(Task.Config(filepath=temp_filepath)).execute(Task.Inputs())
            self.assertEqual(outputs.strs, [])
            self.assertEqual(outputs.num_strs, 0)

            # empty file with newline
            temp_filepath.write_text('\n')
            outputs = Task(Task.Config(filepath=temp_filepath)).execute(Task.Inputs())
            self.assertEqual(outputs.strs, [])
            self.assertEqual(outputs.num_strs, 0)

    def test_not_found(self):
        with tempfile.TemporaryDirectory() as temp_dir:
            temp_dir = pathlib.Path(temp_dir)

            with self.assertRaises(FileNotFoundError):
                Task(Task.Config(filepath=temp_dir / 'not_found.txt')).execute(Task.Inputs())

            outputs = Task(Task.Config(filepath=temp_dir / 'not_found.txt', allow_not_found=True)).execute(Task.Inputs())
            self.assertEqual(outputs.strs, [])
            self.assertEqual(outputs.num_strs, 0)



import unittest
from irisml.tasks.get_key_and_int_list_from_json_string import Task


class TestGetKeyAndIntListFromJsonString(unittest.TestCase):
    def test_simple(self):
        json_string = '{"key1": [1, 2, 3], "key2": [4, 5, 6]}'
        outputs = Task(Task.Config()).execute(Task.Inputs(json_string))
        self.assertEqual(outputs.key_names, ['key1', 'key2'])
        self.assertEqual(outputs.ints[0].tolist(), [1, 2, 3])
        self.assertEqual(outputs.ints[1].tolist(), [4, 5, 6])



import dataclasses
import logging
import pathlib
import torch
import irisml.core

logger = logging.getLogger(__name__)


class Task(irisml.core.TaskBase):
    """Save an offline version of a pytorch model. torch.jit.save()

    The input model must not make any calls to native python functions.

    Config:
        path (Path): Path to save the model to.
    """
    VERSION = '0.1.0'
    CACHE_ENABLED = False

    @dataclasses.dataclass
    class Inputs:
        model: torch.nn.Module

    @dataclasses.dataclass
    class Config:
        path: pathlib.Path

    def execute(self, inputs):
        logger.info(f"Saving model to {self.config.path}")
        torch.jit.save(inputs.model, self.config.path)
        return self.Outputs()



import dataclasses
import logging
import sys
import subprocess
import irisml.core

logger = logging.getLogger(__name__)


class Task(irisml.core.TaskBase):
    """Print various environment information to stdout/stderr.

    This task is designed for Linux environment.

    Currently this task supports:
        - pip list
        - nvidia-smi
        - cat /proc/cpuinfo
        - free

    Config:
        show_pip_info (bool): If True, run `pip list` and print the result.
        show_cuda_info (bool): If True, run `nvidia-smi` and print the result.
        show_cpu_info (bool): If True, run `cat /proc/cpuinfo` and print the result.
        show_memory_info (bool): If True, run `free` and print the result.
    """
    VERSION = '0.1.1'
    CACHE_ENABLED = False

    @dataclasses.dataclass
    class Config:
        show_pip_info: bool = True
        show_cuda_info: bool = True
        show_cpu_info: bool = True
        show_memory_info: bool = True

    def execute(self, inputs):
        if self.config.show_pip_info:
            self._run_command(['pip', 'list'])

        if self.config.show_cuda_info:
            self._run_command(['nvidia-smi'])

        if self.config.show_cpu_info:
            self._run_command(['cat', '/proc/cpuinfo'])

        if self.config.show_memory_info:
            self._run_command(['free'])

        sys.stdout.flush()
        return self.Outputs()

    def dry_run(self, inputs):
        return self.execute(inputs)

    @staticmethod
    def _run_command(args):
        try:
            c = subprocess.run(args, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, text=True)
            print(f"***** {' '.join(args)} *****")
            print(c.stdout)
            if c.returncode != 0:
                logger.warning(f"{args} returned non-zero exit status {c.returncode}")
        except Exception as e:
            logger.warning(f"Failed to run a command {args}. {e}")



import dataclasses
import logging
import pathlib
import typing
import irisml.core

logger = logging.getLogger(__name__)


class Task(irisml.core.TaskBase):
    """Save a list of strings to a text file.

    If the input is None, the task will be skipped.

    Config:
        path (Path): Path to the output file.
    """
    VERSION = '0.1.1'
    CACHE_ENABLED = False

    @dataclasses.dataclass
    class Inputs:
        strs: typing.Optional[typing.List[str]]

    @dataclasses.dataclass
    class Config:
        path: pathlib.Path

    def execute(self, inputs):
        if inputs.strs is None:
            logger.info("inputs.strs is None, skipping")
            return self.Outputs()

        self.config.path.parent.mkdir(parents=True, exist_ok=True)
        self.config.path.write_text('\n'.join(inputs.strs))
        logger.info(f"Saved {len(inputs.strs)} items to {self.config.path}")
        return self.Outputs()



import unittest
import PIL.Image
import torch.utils.data
from irisml.tasks.get_fake_visual_question_answering_dataset import Task


class TestGetFakeImageClassificationDataset(unittest.TestCase):
    def test_simple(self):
        outputs = Task(Task.Config(num_images=100)).execute(Task.Inputs())
        dataset = outputs.dataset
        self.assertIsInstance(dataset, torch.utils.data.Dataset)
        self.assertEqual(len(dataset), 100)
        self.assertEqual(len(dataset[0]), 2)
        self.assertIsInstance(dataset[0][0][0], str)
        self.assertIsInstance(dataset[0][0][1], PIL.Image.Image)
        self.assertIsInstance(dataset[0][1], str)



import copy
import dataclasses
import logging
import typing
import irisml.core

logger = logging.getLogger(__name__)


class Task(irisml.core.TaskBase):
    """Repeat the given tasks for multiple times.

    The tasks are run on a cloned context. The outputs of a run won't be available on the next run. The tasks can use REPEAT_TASKS_INDEX environment variable to get the index of the current run.

    If float_output_names is specified, the task collects those outputs and returns a list.

    Config:
        num_repeats (int): The number of repeats.
        tasks (List[TaskDescription]): The tasks to be repeated.
        float_output_names (List[str]): The name of outputs to collect.
    """
    VERSION = '0.1.0'
    CACHE_ENABLED = False
    RESOLVE_CONFIG_VARIABLES = False

    @dataclasses.dataclass
    class Inputs:
        pass

    @dataclasses.dataclass
    class Config:
        num_repeats: int
        tasks: typing.List[irisml.core.TaskDescription]
        float_output_names: typing.List[str] = dataclasses.field(default_factory=list)

    @dataclasses.dataclass
    class Outputs:
        float_output_values: typing.List[typing.List[float]]

    def execute(self, inputs):
        return self._execute(inputs, dry_run=False)

    def dry_run(self, inputs):
        return self._execute(inputs, dry_run=True)

    def _execute(self, inputs, dry_run):
        config_without_tasks = copy.deepcopy(self.config)
        config_without_tasks.tasks = []
        resolved_config = self.context.resolve(config_without_tasks)
        logger.debug(f"{resolved_config=}")

        for output_names in resolved_config.float_output_names:
            if output_names.count('.') != 1:
                raise ValueError(f"The output name must be '<task_name>.<output_name>'. Actual: {output_names}")

        tasks = []
        for task_description in self.config.tasks:
            task = irisml.core.Task(task_description)
            task.load_module()
            tasks.append(task)

        output_task_variable_names = [names.split('.') for names in resolved_config.float_output_names]

        all_output_values = []
        for index in range(resolved_config.num_repeats):
            logger.info(f"Repeat {index}: Start running the tasks.")

            context = self.context.clone()
            context.add_environment_variable('REPEAT_TASKS_INDEX', index)

            for task in tasks:
                task.execute(context, dry_run=dry_run)

            output_values = [getattr(context.get_outputs(t), v) for t, v in output_task_variable_names]
            all_output_values.append(output_values)

            logger.debug(f"Repeat {index}: The outputs are {output_values}")

        return self.Outputs(all_output_values)



import unittest
import pickle
import torch
from irisml.tasks.pickling_object import Task


class TestPicklingObject(unittest.TestCase):
    def test_tensor(self):
        tensor = torch.rand(3, 4, 5)
        outputs = Task(Task.Config()).execute(Task.Inputs(tensor))
        self.assertTrue(torch.equal(pickle.loads(outputs.data), tensor))

    def test_str(self):
        string = "Hello World"
        outputs = Task(Task.Config()).execute(Task.Inputs(string))
        self.assertEqual(pickle.loads(outputs.data), string)

    def test_list_of_tensors(self):
        list_of_tensors = [torch.Tensor([[0, 1, 0, 0, 1, 1]]), torch.Tensor([[1, 1, 0, 0, 1, 1]])]
        outputs = Task(Task.Config()).execute(Task.Inputs(list_of_tensors))
        self.assertTrue(all([torch.equal(a, b) for a, b in zip(pickle.loads(outputs.data), list_of_tensors)]))



import dataclasses
import logging
import pathlib
import typing
import torch
import irisml.core

logger = logging.getLogger(__name__)


class Task(irisml.core.TaskBase):
    """Save images from a dataset to disk.

    Config:
        dirpath (pathlib.Path): Directory to save images to.
        extension (str): File extension to use for images.
        indices (torch.Tensor): Indices of images to save. If None, all images are saved.
    """
    VERSION = '0.2.0'
    CACHE_ENABLED = False

    @dataclasses.dataclass
    class Inputs:
        dataset: torch.utils.data.Dataset

    @dataclasses.dataclass
    class Config:
        dirpath: pathlib.Path = pathlib.Path('.')
        extension: str = 'png'
        indices: typing.Optional[torch.Tensor] = None

    def execute(self, inputs):
        self.config.dirpath.mkdir(parents=True, exist_ok=True)
        if self.config.indices is None:
            for i, (image, targets) in enumerate(inputs.dataset):
                self._save_image(image, targets, i)
            logger.info(f"Saved {len(inputs.dataset)} images to {self.config.dirpath}.")
        else:
            if self.config.indices.dim() != 1:
                raise ValueError(f'Indices must be a 1D tensor, got {self.config.indices.dim()}D.')

            for index in self.config.indices.tolist():
                image, targets = inputs.dataset[index]
                self._save_image(image, targets, index)
            logger.info(f"Saved {len(self.config.indices)} images to {self.config.dirpath}.")

        return self.Outputs()

    def _save_image(self, image, targets, index):
        filepath = self.config.dirpath / f'{index}.{self.config.extension}'
        image.save(filepath)
        logger.info(f'Index {index}: Saved image to {filepath}, targets: {targets}')



import dataclasses
import logging
import typing
import irisml.core

logger = logging.getLogger(__name__)


class Task(irisml.core.TaskBase):
    """Returns one of the inputs based on a condition.

    If the condition is true, the task returns then_* inputs, otherwise it returns else_* inputs.
    """

    VERSION = '0.1.0'

    @dataclasses.dataclass
    class Inputs:
        condition: bool
        then_float: typing.Optional[float] = None
        then_int: typing.Optional[int] = None
        then_str: typing.Optional[str] = None
        then_str_list: typing.Optional[typing.List[str]] = None
        else_float: typing.Optional[float] = None
        else_int: typing.Optional[int] = None
        else_str: typing.Optional[str] = None
        else_str_list: typing.Optional[typing.List[str]] = None

    @dataclasses.dataclass
    class Outputs:
        result_float: typing.Optional[float] = None
        result_int: typing.Optional[int] = None
        result_str: typing.Optional[str] = None
        result_str_list: typing.Optional[typing.List[str]] = None

    def execute(self, inputs):
        type_names = ['float', 'int', 'str', 'str_list']
        then_inputs = [inputs.then_float, inputs.then_int, inputs.then_str, inputs.then_str_list]
        else_inputs = [inputs.else_float, inputs.else_int, inputs.else_str, inputs.else_str_list]

        if sum(1 for x in then_inputs if x is not None) != 1:
            raise ValueError(f"Expected exactly one non-None element in then_inputs, got {then_inputs}")
        if sum(1 for x in else_inputs if x is not None) != 1:
            raise ValueError(f"Expected exactly one non-None element in else_inputs, got {else_inputs}")

        index = next(i for i, x in enumerate(then_inputs) if x is not None)
        if else_inputs[index] is None:
            raise ValueError(f"Expected else_{type_names[index]} to be non-None, got {else_inputs[index]}")

        logger.debug(f"then_{type_names[index]}, else_{type_names[index]}: {then_inputs[index]}, {else_inputs[index]}. Condition: {inputs.condition}")
        if inputs.condition:
            result = then_inputs[index]
        else:
            result = else_inputs[index]

        return self.Outputs(**{f'result_{type_names[index]}': result})

    def dry_run(self, inputs):
        return self.execute(inputs)



import dataclasses
import logging
import random
import irisml.core
import torch.utils.data
import PIL.Image
import PIL.ImageColor

logger = logging.getLogger(__name__)


class Task(irisml.core.TaskBase):
    """Generate a fake visual question answering dataset.

    A generated dataset returns ((question: str, image: PIL.Image), answer: str) pairs.

    Images are 224x224 pixels, and are filled with a single color. The color is chosen randomly from the list of
    colors in PIL.ImageColor.colormap, and the list is shuffled with the given random seed.

    Config:
        num_images (int): Number of images to generate. Default: 100.
        question (str): Question to ask about each image. Default: 'What is the color of this image? <|image|>'.
        random_seed (int): Random seed. Default: 0.
    """
    VERSION = '0.1.0'

    @dataclasses.dataclass
    class Config:
        num_images: int = 100
        question: str = 'What is the color of this image? <|image|>'
        random_seed: int = 0

    @dataclasses.dataclass
    class Outputs:
        dataset: torch.utils.data.Dataset

    class FakeDataset(torch.utils.data.Dataset):
        def __init__(self, num_images, random_seed, question):
            color_names = list(PIL.ImageColor.colormap.keys())
            rng = random.Random(random_seed)
            rng.shuffle(color_names)
            logger.debug(f"Generated color names: {color_names} (random_seed={random_seed})")

            self._color_names = color_names
            self._num_images = num_images
            self._question = question

        def __len__(self):
            return self._num_images

        def __getitem__(self, index):
            if index >= self._num_images:
                raise IndexError(f"Index out of range: {index} >= {self._num_images}")

            color_name = self._color_names[index % len(self._color_names)]
            image = PIL.Image.new('RGB', (224, 224), color=color_name)
            return (self._question, image), color_name

    def execute(self, inputs):
        dataset = Task.FakeDataset(self.config.num_images, self.config.random_seed, self.config.question)
        return self.Outputs(dataset)

    def dry_run(self, inputs):
        return self.execute(inputs)



import pathlib
import tempfile
import unittest
import torch
from irisml.tasks.load_float_tensor_jsonl import Task


class TestLoadFloatTensorJsonl(unittest.TestCase):
    def test_simple(self):
        with tempfile.NamedTemporaryFile() as f:
            path = pathlib.Path(f.name)
            path.write_text('{"key": [1.0, 2.0, 3.0]}\n{"key": [4.0, 5.0, 6.0]}\n')
            outputs = Task(Task.Config(path=path, key_name='key')).execute(Task.Inputs())
            self.assertIsInstance(outputs.data, torch.Tensor)
            self.assertEqual(outputs.data.shape, (2, 3))
            self.assertEqual(outputs.data.tolist(), [[1.0, 2.0, 3.0], [4.0, 5.0, 6.0]])



import unittest
import unittest.mock
from irisml.tasks.get_secret_from_azure_keyvault import Task


class TestGetSecretFromAzureKeyvault(unittest.TestCase):
    @unittest.mock.patch('irisml.tasks.get_secret_from_azure_keyvault.SecretClient')
    def test_simple(self, mock_client):
        mock_secret = unittest.mock.MagicMock(value='secret_value')
        mock_client.return_value.get_secret.return_value = mock_secret
        outputs = Task(Task.Config('example', 'secret_name')).execute(Task.Inputs())
        self.assertEqual(outputs.secret, 'secret_value')
        mock_client.assert_called_with(vault_url='https://example.vault.azure.net', credential=unittest.mock.ANY)



import typing
import unittest
import torch
from irisml.core import Context
from irisml.tasks.get_dataset_subset import Task


class DummyDataset(torch.utils.data.Dataset):
    def __init__(self, data: typing.List):
        self._data = data

    def __len__(self):
        return len(self._data)

    def __getitem__(self, x):
        return self._data[x]


class TestGetDatasetSubset(unittest.TestCase):
    def test_simple(self):
        dataset = DummyDataset(list(range(100)))

        inputs = Task.Inputs(dataset=dataset)
        config = Task.Config(num_images=10)

        context = Context()
        task = Task(config, context)
        outputs = task.execute(inputs)
        self.assertEqual(len(outputs.dataset), 10)



import unittest
from irisml.tasks.if_else import Task


class TestIfElse(unittest.TestCase):
    def test_simple(self):
        outputs = Task(Task.Config()).execute(Task.Inputs(condition=True, then_float=1.0, else_float=2.0))
        self.assertEqual(outputs.result_float, 1.0)
        self.assertIsNone(outputs.result_int)
        self.assertIsNone(outputs.result_str)
        self.assertIsNone(outputs.result_str_list)

        outputs = Task(Task.Config()).execute(Task.Inputs(condition=False, then_float=1.0, else_float=2.0))
        self.assertEqual(outputs.result_float, 2.0)

        outputs = Task(Task.Config()).execute(Task.Inputs(condition=True, then_str_list=['a', 'b'], else_str_list=['c', 'd']))
        self.assertEqual(outputs.result_str_list, ['a', 'b'])
        self.assertIsNone(outputs.result_float)
        self.assertIsNone(outputs.result_int)
        self.assertIsNone(outputs.result_str)

    def test_invalid(self):
        with self.assertRaises(ValueError):
            Task(Task.Config()).execute(Task.Inputs(condition=True))

        with self.assertRaises(ValueError):
            Task(Task.Config()).execute(Task.Inputs(condition=True, then_float=1.0, then_int=2))

        with self.assertRaises(ValueError):
            Task(Task.Config()).execute(Task.Inputs(condition=True, then_float=1.0, then_str='a', else_float=2.0, else_str='b'))



import dataclasses
import logging
import pathlib
import irisml.core

logger = logging.getLogger(__name__)


class Task(irisml.core.TaskBase):
    """Save the given input binary to a file."""
    VERSION = '0.1.0'
    CACHE_ENABLED = False

    @dataclasses.dataclass
    class Inputs:
        data: bytes

    @dataclasses.dataclass
    class Config:
        path: pathlib.Path

    def execute(self, inputs):
        if self.config.path.exists():
            logger.warning(f"Path {self.config.path} already exists. Aborting.")
        else:
            self.config.path.parent.mkdir(parents=True, exist_ok=True)
            self.config.path.write_bytes(inputs.data)
        return self.Outputs()



import dataclasses
import logging
import random
import typing
import irisml.core
import torch
import PIL.Image
import PIL.ImageDraw

logger = logging.getLogger(__name__)


class Task(irisml.core.TaskBase):
    """Generate a fake object detection dataset.

    A generated dataset returns (image:PIL.Image, targets:List[Tensor[N, 5]])

    Configs:
        num_images (int): The number of images to make. Default: 100
        num_classes (int): The number of classes. Default: 10
        num_max_boxes (int): The max number of objects per image. Default: 10
        image_sizes ([(int, int), ...]): A list of candidate image sizes. Default: [(320, 320)]
        random_seed (int): The random seed. Default: 0
        targets ([[[float, float, float, float, float], ...], ...]): A list of targets. If not specified, random targets are generated. Default: None
    """
    VERSION = '0.1.5'

    @dataclasses.dataclass
    class Config:
        num_images: int = 100
        num_classes: int = 10
        num_max_boxes: int = 10  # The max number of objects per image.
        image_sizes: typing.List[typing.Tuple[int, int]] = dataclasses.field(default_factory=lambda: [(320, 320)])
        random_seed: int = 0
        targets: typing.Optional[typing.List[typing.List[typing.List[float]]]] = None

    @dataclasses.dataclass
    class Outputs:
        dataset: torch.utils.data.Dataset
        num_classes: int
        class_names: typing.List[str]

    class FakeDataset(torch.utils.data.Dataset):
        def __init__(self, targets: typing.List[torch.Tensor], colors: typing.List[typing.Tuple[int, int, int]], sizes: typing.List[typing.Tuple[int, int]]):
            assert len(targets) == len(sizes)
            self._targets = targets
            self._colors = colors
            self._sizes = sizes

        def __len__(self):
            return len(self._targets)

        def __getitem__(self, index):
            targets = self._targets[index]
            return self._generate_image(index, targets), targets

        def _generate_image(self, index, targets):
            size = self._sizes[index]
            image = PIL.Image.new('RGB', size)
            draw = PIL.ImageDraw.Draw(image)
            for t in targets:
                rect = (t[1] * size[0], t[2] * size[1], t[3] * size[0], t[4] * size[1])
                draw.rectangle(rect, fill=self._colors[int(t[0])])
            return image

    def execute(self, inputs):
        rng = random.Random(self.config.random_seed)

        def _random_box(image_width, image_height):
            # Make sure the bounding box size is at least 1 pixel.
            cx = rng.uniform(1 / image_width / 2, 1 - 1 / image_width / 2)
            cy = rng.uniform(1 / image_height / 2, 1 - 1 / image_height / 2)
            w = rng.uniform(1 / image_width, min(1 - cx, cx) * 2)
            h = rng.uniform(1 / image_height, min(1 - cy, cy) * 2)
            return [cx - w / 2, cy - h / 2, cx + w / 2, cy + h / 2]

        if not self.config.image_sizes:
            raise ValueError("Image sizes setting is empty.")

        for c in self.config.image_sizes:
            if len(c) != 2 or c[0] <= 0 or c[1] <= 0:
                raise ValueError(f"Invalid image size configs: {self.config.image_sizes}")

        sizes = rng.choices(self.config.image_sizes, k=self.config.num_images)

        if self.config.targets:
            targets = [torch.tensor(t) for t in self.config.targets]
            if len(targets) != self.config.num_images:
                raise ValueError(f"Invalid targets: {self.config.targets}")
        else:
            targets = [torch.tensor([[rng.randrange(self.config.num_classes), *_random_box(*sizes[i])] for _ in range(rng.randrange(self.config.num_max_boxes + 1))]).reshape(-1, 5)
                       for i in range(self.config.num_images)]

        colors = [(rng.randint(0, 255), rng.randint(0, 255), rng.randint(0, 255)) for _ in range(self.config.num_classes)]
        logger.debug(f"Generated labels: {colors}")

        dataset = Task.FakeDataset(targets, colors, sizes)
        class_names = [f'class_{i}' for i in range(self.config.num_classes)]
        return self.Outputs(dataset, num_classes=self.config.num_classes, class_names=class_names)

    def dry_run(self, inputs):
        return self.execute(inputs)



import pathlib
import tempfile
import unittest
from irisml.tasks.save_strs_text import Task


class TestSaveStrsText(unittest.TestCase):
    def test_simple(self):
        with tempfile.TemporaryDirectory() as temp_dir:
            temp_dir = pathlib.Path(temp_dir)
            Task(Task.Config(path=temp_dir / 'new_dir' / 'new_file.txt')).execute(Task.Inputs(['hello', 'world']))
            self.assertTrue((temp_dir / 'new_dir' / 'new_file.txt').exists())
            self.assertEqual((temp_dir / 'new_dir' / 'new_file.txt').read_text(), 'hello\nworld')

    def test_none_inputs(self):
        with tempfile.TemporaryDirectory() as temp_dir:
            temp_path = pathlib.Path(temp_dir) / 'new_dir' / 'new_file.txt'
            Task(Task.Config(path=temp_path)).execute(Task.Inputs(None))
            self.assertFalse(temp_path.exists())



import collections
import dataclasses
import logging
import math
import random
from typing import Optional
import torch.utils.data
import irisml.core

logger = logging.getLogger(__name__)


class Task(irisml.core.TaskBase):
    """Get a train/val split of a dataset.

    If val_dataset is not provided, then the train_dataset will be split into a train and val dataset. E.g. if train_val_split is 0.8, then 80% of the
    train_dataset will be used for training and 20% will be used for validation.

    If val_dataset is provided, then it will be used as the validation dataset and the train_dataset will be used for training as-is.

    If the dataset instance has 'get_targets(index)' method, it will be used for getting targets.

    Config:
        train_val_split (float): The ratio of the size of the train/val datasets.
        keep_class_distribution (bool): If true, tries to keep the class distribution.
    """
    VERSION = '0.2.1'

    @dataclasses.dataclass
    class Config:
        train_val_split: float
        keep_class_distribution: bool = False

    @dataclasses.dataclass
    class Inputs:
        train_dataset: torch.utils.data.Dataset
        val_dataset: Optional[torch.utils.data.Dataset] = None

    @dataclasses.dataclass
    class Outputs:
        train_dataset: torch.utils.data.Dataset
        val_dataset: torch.utils.data.Dataset

    def execute(self, inputs):
        if not 0 <= self.config.train_val_split <= 1:
            raise ValueError("train_val_split must be between 0 and 1")

        if not inputs.val_dataset:
            if self.config.keep_class_distribution:
                train_indices, val_indices = self._get_split_indices(inputs.train_dataset, self.config.train_val_split)
                assert len(train_indices) + len(val_indices) == len(inputs.train_dataset)
                train_dataset = torch.utils.data.Subset(inputs.train_dataset, train_indices)
                val_dataset = torch.utils.data.Subset(inputs.train_dataset, val_indices)
            else:
                num_images = len(inputs.train_dataset)
                num_train_images = int(num_images * self.config.train_val_split)
                num_val_images = num_images - num_train_images
                train_dataset, val_dataset = torch.utils.data.random_split(inputs.train_dataset, [num_train_images, num_val_images])
            logger.info(f"Split train dataset into {len(train_dataset)} train images and {len(val_dataset)} val images")
        else:
            train_dataset, val_dataset = inputs.train_dataset, inputs.val_dataset
            logger.info(f"Skip splitting - val dataset is already provided. Using {len(train_dataset)} train images and {len(val_dataset)} val images")

        return self.Outputs(train_dataset, val_dataset)

    def dry_run(self, inputs):
        return self.execute(inputs)

    @staticmethod
    def _get_split_indices(dataset, train_val_ratio):
        """Get train/val indices that keeps the class distribution."""
        def get_class(x):
            if torch.is_tensor(x):
                if not x.shape:
                    return [int(x)]
                elif len(x.shape) == 1:
                    return x.tolist()
                elif len(x.shape) == 2:
                    assert x.shape[1] == 5
                    return [int(y[0]) for y in x]
            elif isinstance(x, int):
                return [x]
            raise ValueError(f"Unexpected data label: {x=}, {type(x)=}")

        if hasattr(dataset, 'get_targets'):
            targets = [dataset.get_targets(i) for i in range(len(dataset))]
        else:
            targets = [x[1] for x in dataset]

        image_class_indices = [(i, get_class(t)) for i, t in enumerate(targets)]
        random.shuffle(image_class_indices)

        train_indices = []
        val_indices = []

        # Split examples that has no class labels.
        negative_examples = [i for i, c in image_class_indices if not c]
        num_val_negative_examples = math.ceil(len(negative_examples) * (1 - train_val_ratio))
        val_indices += negative_examples[0:num_val_negative_examples]
        train_indices += negative_examples[num_val_negative_examples:]

        image_class_indices = [(i, c) for i, c in image_class_indices if c]
        train_classes_counter = collections.Counter()
        val_classes_counter = collections.Counter()
        for index, classes in image_class_indices:
            train_class_counts = [train_classes_counter[x] for x in classes]
            val_class_counts = [val_classes_counter[x] for x in classes]
            train_class_count_sum = sum(train_class_counts) * (1 - train_val_ratio) / train_val_ratio
            train_class_count_min = min(train_class_counts) * (1 - train_val_ratio) / train_val_ratio
            val_class_count_sum = sum(val_class_counts)
            val_class_count_min = min(val_class_counts)
            if val_class_count_min < train_class_count_min or (val_class_count_min == train_class_count_min and val_class_count_sum < train_class_count_sum):
                val_indices.append(index)
                val_classes_counter.update(classes)
            else:
                train_indices.append(index)
                train_classes_counter.update(classes)
        return train_indices, val_indices



import dataclasses
import json
import logging
import pathlib
import torch
import irisml.core

logger = logging.getLogger(__name__)


class Task(irisml.core.TaskBase):
    """Save a 2D float tensor to a JSONL file.

    The input tensor is expected to be a 2D float tensor. Each row of the tensor
    is converted to a JSON object and written to the output file.

    Config:
        path (Path): Path to the output file.
        key_name (str): Name of the key to use for each item in the list.
    """
    VERSION = '0.1.1'
    CACHE_ENABLED = False

    @dataclasses.dataclass
    class Inputs:
        data: torch.Tensor

    @dataclasses.dataclass
    class Config:
        path: pathlib.Path
        key_name: str

    def execute(self, inputs):
        if inputs.data.dim() != 2:
            raise ValueError(f"Expected 2D tensor, got {inputs.data.dim()}D")

        self.config.path.parent.mkdir(parents=True, exist_ok=True)
        with self.config.path.open('w') as f:
            for item in inputs.data:
                json.dump({self.config.key_name: item.tolist()}, f)
                f.write('\n')

        logger.info(f"Saved {len(inputs.data)} items to {self.config.path}")
        return self.Outputs()



import dataclasses
import json
import logging
import pathlib
import torch
import irisml.core

logger = logging.getLogger(__name__)


class Task(irisml.core.TaskBase):
    """Load a 2D float tensor from a JSONL file.

    This task is paired with the `save_float_tensor_jsonl` task.

    Config:
        path (Path): Path to the JSONL file.
        key_name (str): Name of the key to extract from each JSON object.
    """
    VERSION = '0.1.0'

    @dataclasses.dataclass
    class Config:
        path: pathlib.Path
        key_name: str

    @dataclasses.dataclass
    class Outputs:
        data: torch.Tensor

    def execute(self, inputs):
        logger.info(f"Loading {self.config.path}")
        data = []
        with open(self.config.path) as f:
            for line in f:
                if not line.strip():
                    continue
                data.append(json.loads(line)[self.config.key_name])

        data = torch.tensor(data)
        logger.info(f"Loaded {len(data)} lines. Shape: {data.shape}")

        return self.Outputs(data=data)

    def dry_run(self, inputs):
        return self.Outputs(data=torch.zeros(0, 1))



import dataclasses
import json
import logging
import pathlib
import typing
import irisml.core

logger = logging.getLogger(__name__)


class Task(irisml.core.TaskBase):
    """Load a list of strings from a JSONL file.

    Config:
        path (Path): Path to the JSONL file.
        key_name (str): Name of the key to extract from each JSON object.
    """
    VERSION = '0.1.0'

    @dataclasses.dataclass
    class Config:
        path: pathlib.Path
        key_name: str

    @dataclasses.dataclass
    class Outputs:
        data: typing.List[str]

    def execute(self, inputs):
        logger.info(f"Loading {self.config.path}")
        data = []
        with open(self.config.path) as f:
            for line in f:
                if not line.strip():
                    continue
                data.append(json.loads(line)[self.config.key_name])

        logger.info(f"Loaded {len(data)} lines")
        return self.Outputs(data=data)

    def dry_run(self, inputs):
        return self.Outputs(data=["dry_run"])



import dataclasses
import logging
import typing
import irisml.core


PLACEHOLDER = '<|placeholder|>'
PLACEHOLDER2 = '<|placeholder2|>'
logger = logging.getLogger(__name__)


class Task(irisml.core.TaskBase):
    """Make a prompt for each string.

    For example, if the template is "What is <|placeholder|>?" and the strings are ["a", "b", "c"], the prompts will be ["What is a?", "What is b?", "What is c?"].

    Optionally, you can provide a second list of strings. If you do, the template must contain "<|placeholder2|>" and the two lists must be the same length.
    The second list of strings will be used to replace "<|placeholder2|>" in the prompts.

    Config:
        template (str): The template to use for the prompt. Must contain "<|placeholder|>".
    """
    VERSION = '0.2.0'

    @dataclasses.dataclass
    class Inputs:
        strings: typing.List[str]
        strings2: typing.Optional[typing.List[str]] = None

    @dataclasses.dataclass
    class Config:
        template: str

    @dataclasses.dataclass
    class Outputs:
        prompts: typing.List[str]

    def execute(self, inputs):
        if PLACEHOLDER not in self.config.template:
            raise ValueError(f'"{PLACEHOLDER}" must be in template')

        if inputs.strings2 is not None:
            if len(inputs.strings) != len(inputs.strings2):
                raise ValueError('len(strings) != len(strings2)')
            if PLACEHOLDER2 not in self.config.template:
                raise ValueError(f'"{PLACEHOLDER2}" must be in template')
        else:
            if PLACEHOLDER2 in self.config.template:
                raise ValueError(f'strings2 must be provided if "{PLACEHOLDER2}" is in template')

        prompts = [self.config.template.replace(PLACEHOLDER, s) for s in inputs.strings]
        if inputs.strings2 is not None:
            prompts = [p.replace(PLACEHOLDER2, s) for p, s in zip(prompts, inputs.strings2)]

        for p in prompts:
            logger.info(f"Created a prompt: {p}")

        return self.Outputs(prompts=prompts)

    def dry_run(self, inputs):
        return self.execute(inputs)



import unittest
import torch
from irisml.tasks.get_fake_image_classification_multilabel_dataset import Task


class TestGetFakeImageClassificationMultilabelDataset(unittest.TestCase):
    def test_simple(self):
        outputs = Task(Task.Config(num_images=100, num_classes=5)).execute(Task.Inputs())
        self.assertEqual(len(outputs.dataset), 100)
        self.assertEqual(outputs.num_classes, 5)
        self.assertEqual(len(outputs.class_names), 5)

        first_image, first_target = outputs.dataset[0]
        self.assertEqual(first_image.size, (224, 224))
        self.assertEqual(first_target.size(), (5,))
        self.assertEqual(first_target.dtype, torch.int)
        self.assertTrue(x in [0, 1] for x in first_target)



import dataclasses
import typing
import irisml.core


class Task(irisml.core.TaskBase):
    """Make a text transform function.

    This task creates a text transform function that replaces "{}" in a template with a given string.

    Config:
        template (str): The template to use for the text transform function. Must contain "{}".
    """
    VERSION = '0.1.0'

    @dataclasses.dataclass
    class Config:
        template: str

    @dataclasses.dataclass
    class Outputs:
        transform: typing.Callable[[str], str]

    def execute(self, inputs):
        return self.Outputs(TextTransform(self.config.template))

    def dry_run(self, inputs):
        return self.execute(inputs)


class TextTransform:
    def __init__(self, template):
        self._template = template
        if '{}' not in self._template:
            raise ValueError('Template must contain "{}"')

    def __call__(self, text):
        return self._template.replace('{}', text)



import copy
import dataclasses
import io
import logging
import pathlib
import typing
import torch
import irisml.core

logger = logging.getLogger(__name__)


class Task(irisml.core.TaskBase):
    """Load a state_dict from various sources.

    Supported sources are:
        1. a dictionary object. Use Inputs.state_dict.
        2. a bytes object. Use Inputs.state_dict_bytes.
        3. a file on local filesystem. Use Config.path.

    Config:
        path (pathlib.Path): [optional] A file path to the state_dict.
        strict (bool): Load state_dict in strict mode
        ignore_shape_mismatch (bool): If True, ignore weights with a wrong shape. 'strict' must be False.
    """
    VERSION = '0.1.1'

    @dataclasses.dataclass
    class Inputs:
        model: torch.nn.Module
        state_dict: typing.Optional[typing.Dict] = None
        state_dict_bytes: typing.Optional[bytes] = None

    @dataclasses.dataclass
    class Config:
        path: typing.Optional[pathlib.Path] = None
        strict: bool = True
        ignore_shape_mismatch: bool = False

    @dataclasses.dataclass
    class Outputs:
        model: torch.nn.Module

    def execute(self, inputs: Inputs):
        num_sources = len([x for x in [inputs.state_dict, inputs.state_dict_bytes, self.config.path] if x])
        if num_sources != 1:
            raise ValueError("One state_dict source must be provided.")

        if inputs.state_dict:
            state_dict = inputs.state_dict
        if inputs.state_dict_bytes:
            state_dict = torch.load(io.BytesIO(inputs.state_dict_bytes), map_location='cpu')
        elif self.config.path:
            state_dict = torch.load(self._config.path, map_location='cpu')

        if not state_dict:
            raise ValueError("Failed to load the state_dict.")

        if self.config.ignore_shape_mismatch:
            # Filter parameters with different shape.
            existing_state_dict = inputs.model.state_dict()
            new_state_dict = {k: v for k, v in state_dict.items() if k not in existing_state_dict or v.shape == existing_state_dict[k].shape}

            ignored_keys = set(state_dict) - set(new_state_dict)
            for key in ignored_keys:
                logger.info(f"Ignored {key} in the state_dict because it has unexpected shape. Actual: {state_dict[key].shape}, Expected: {existing_state_dict[key].shape}")

            state_dict = new_state_dict

        model = copy.deepcopy(inputs.model)
        model.load_state_dict(state_dict, strict=self.config.strict)
        return self.Outputs(model)

    def dry_run(self, inputs):
        return self.Outputs(inputs.model)



import unittest
import PIL.Image
import torch
from irisml.tasks.get_fake_image_text_classification_dataset import Task


class TestGetFakeImageTextClassificationDataset(unittest.TestCase):
    def test_simple(self):
        outputs = Task(Task.Config(num_images=10, num_classes=2)).execute(Task.Inputs())

        self.assertEqual(len(outputs.dataset), 10)
        self.assertEqual(len(outputs.dataset[0]), 2)
        self.assertIsInstance(outputs.dataset[0][0][0], PIL.Image.Image)
        self.assertIsInstance(outputs.dataset[0][0][1], str)
        self.assertIsInstance(outputs.dataset[0][1], torch.Tensor)
        self.assertEqual(outputs.dataset[0][1].ndim, 0)

        self.assertEqual(outputs.num_classes, 2)
        self.assertEqual(len(outputs.class_names), 2)



import io
import pathlib
import tempfile
import unittest

import torch
from irisml.tasks.load_state_dict import Task


class TestLoadStateDict(unittest.TestCase):
    def test_simple(self):
        fake_model = torch.nn.Conv2d(3, 3, 3)
        fake_model2 = torch.nn.Conv2d(3, 3, 3)
        self.assertFalse(self._is_equal_state_dict(fake_model.state_dict(), fake_model2.state_dict()))

        # Load from a file.
        inputs = Task.Inputs(model=fake_model)
        with tempfile.TemporaryDirectory() as temp_dir:
            temp_file = pathlib.Path(temp_dir) / 'test.pth'
            torch.save(fake_model2.state_dict(), temp_file)

            config = Task.Config(path=temp_file)

            task = Task(config)
            outputs = task.execute(inputs)
            model = outputs.model
            self.assertTrue(self._is_equal_state_dict(model.state_dict(), fake_model2.state_dict()))

        # Load from state_dict
        task = Task(Task.Config())
        fake_model = torch.nn.Conv2d(3, 3, 3)
        inputs = Task.Inputs(model=fake_model, state_dict=fake_model2.state_dict())
        outputs = task.execute(inputs)
        self.assertTrue(self._is_equal_state_dict(outputs.model.state_dict(), fake_model2.state_dict()))

        # Load from state_dict_bytes
        state_buffer = io.BytesIO()
        torch.save(fake_model2.state_dict(), state_buffer)
        fake_model = torch.nn.Conv2d(3, 3, 3)
        inputs = Task.Inputs(model=fake_model, state_dict_bytes=state_buffer.getbuffer().tobytes())
        outputs = task.execute(inputs)
        self.assertTrue(self._is_equal_state_dict(outputs.model.state_dict(), fake_model2.state_dict()))

    def test_copied(self):
        fake_model = torch.nn.Conv2d(3, 3, 3)
        fake_model2 = torch.nn.Conv2d(3, 3, 3)
        task = Task(Task.Config())
        inputs = Task.Inputs(model=fake_model, state_dict=fake_model2.state_dict())
        task.execute(inputs)
        self.assertFalse(self._is_equal_state_dict(fake_model.state_dict(), fake_model2.state_dict()))

    def test_multiple_source(self):
        fake_model = torch.nn.Conv2d(3, 3, 3)
        task = Task(Task.Config('temp.path'))
        inputs = Task.Inputs(model=fake_model, state_dict={'a': torch.zeros(1)})
        with self.assertRaises(ValueError):
            task.execute(inputs)

        inputs = Task.Inputs(model=fake_model, state_dict_bytes=b'123')
        with self.assertRaises(ValueError):
            task.execute(inputs)

        task = Task(Task.Config())
        inputs = Task.Inputs(model=fake_model, state_dict={'a': torch.zeros(1)}, state_dict_bytes=b'123')
        with self.assertRaises(ValueError):
            task.execute(inputs)

    def test_different_shape(self):
        fake_model = torch.nn.Conv2d(3, 4, 3)
        fake_model2 = torch.nn.Conv2d(3, 3, 3)

        # It will be ignored if ignore_shape_mismatch=True
        Task(Task.Config(strict=False, ignore_shape_mismatch=True)).execute(Task.Inputs(model=fake_model, state_dict=fake_model2.state_dict()))

        with self.assertRaises(RuntimeError):
            Task(Task.Config(strict=False)).execute(Task.Inputs(model=fake_model, state_dict=fake_model2.state_dict()))

    def _is_equal_state_dict(self, state_dict1, state_dict2):
        if set(state_dict1) != set(state_dict2):
            return False

        for key in state_dict1:
            if not torch.equal(state_dict1[key], state_dict2[key]):
                return False
        return True



import unittest
import torch
from irisml.tasks.make_prompt_list_with_strings import Task


class TestMakePromptListWithStrings(unittest.TestCase):
    def test_simple(self):
        strings = [f'str{i}' for i in range(10)]
        indices = [torch.tensor([0, 1, 2]), torch.tensor([3]), torch.tensor([6, 7, 8])]
        outputs = Task(Task.Config(template='Here is a list. <|placeholder|>')).execute(Task.Inputs(strings=strings, indices=indices))
        self.assertEqual(outputs.prompts, ['Here is a list. str0, str1, str2', 'Here is a list. str3', 'Here is a list. str6, str7, str8'])



import unittest
import torch
from irisml.tasks.get_dataset_stats import Task


class FakeDataset(torch.utils.data.Dataset):
    def __init__(self, data):
        self._data = data

    def __len__(self):
        return len(self._data)

    def __getitem__(self, index):
        return self._data[index]


class TestGetDatasetStats(unittest.TestCase):
    def test_multiclass_classification(self):
        dataset = FakeDataset([(torch.zeros((3, 100, 100)), torch.tensor([0])),
                               (torch.zeros((3, 100, 100)), torch.tensor([1])),
                               (torch.zeros((3, 100, 100)), torch.tensor([2]))])
        task = Task(Task.Config())
        outputs = task.execute(Task.Inputs(dataset))
        self.assertEqual(outputs.num_images, 3)
        self.assertEqual(outputs.num_classes, 3)
        self.assertEqual(outputs.dataset_type, 'multiclass_classification')

    def test_multiclass_classification_scalar(self):
        dataset = FakeDataset([(torch.zeros((3, 100, 100)), torch.tensor(0)),
                               (torch.zeros((3, 100, 100)), torch.tensor(1)),
                               (torch.zeros((3, 100, 100)), torch.tensor(2))])
        task = Task(Task.Config())
        outputs = task.execute(Task.Inputs(dataset))
        self.assertEqual(outputs.num_images, 3)
        self.assertEqual(outputs.num_classes, 3)
        self.assertEqual(outputs.dataset_type, 'multiclass_classification')

    def test_multilabel_classification(self):
        dataset = FakeDataset([(torch.zeros((3, 100, 100)), torch.tensor([0, 1, 2])),
                               (torch.zeros((3, 100, 100)), torch.tensor([0])),
                               (torch.zeros((3, 100, 100)), torch.tensor([]))])
        task = Task(Task.Config())
        outputs = task.execute(Task.Inputs(dataset))
        self.assertEqual(outputs.num_images, 3)
        self.assertEqual(outputs.num_classes, 3)
        self.assertEqual(outputs.dataset_type, 'multilabel_classification')

    def test_object_detection(self):
        dataset = FakeDataset([(torch.zeros((3, 100, 100)), torch.tensor([[0, 0.1, 0.1, 0.2, 0.2], [1, 0.1, 0.1, 0.2, 0.2], [2, 0.1, 0.1, 0.2, 0.2]])),
                               (torch.zeros((3, 100, 100)), torch.tensor([[0, 0.1, 0.1, 0.2, 0.2]])),
                               (torch.zeros((3, 100, 100)), torch.tensor([]))])
        task = Task(Task.Config())
        outputs = task.execute(Task.Inputs(dataset))
        self.assertEqual(outputs.num_images, 3)
        self.assertEqual(outputs.num_classes, 3)
        self.assertEqual(outputs.dataset_type, 'object_detection')



import pathlib
import tempfile
import unittest
from irisml.core import Context
from irisml.tasks.save_file import Task


class TestSaveFile(unittest.TestCase):
    def test_simple(self):
        inputs = Task.Inputs(b'12345')
        with tempfile.TemporaryDirectory() as temp_dir:
            temp_file = pathlib.Path(temp_dir) / 'directory' / 'myfile'
            config = Task.Config(temp_file)
            task = Task(config, Context())
            task.execute(inputs)

            self.assertTrue(temp_file.exists())
            self.assertEqual(temp_file.read_bytes(), b'12345')



import ctypes
import dataclasses
import logging
import typing
import torch
import irisml.core

logger = logging.getLogger(__name__)


class Task(irisml.core.TaskBase):
    """Emulate FP8 quantization.

    Config:
        fp8_format ('e4m3' or 'e5m2'): FP8 format to use.
    """
    VERSION = '0.1.0'

    @dataclasses.dataclass
    class Inputs:
        tensor: torch.Tensor

    @dataclasses.dataclass
    class Config:
        fp8_format: typing.Literal['e4m3', 'e5m2'] = 'e4m3'

    @dataclasses.dataclass
    class Outputs:
        tensor: torch.Tensor

    def execute(self, inputs):
        if self.config.fp8_format == 'e4m3':
            fp8_num_fraction_bits = 3
            fp8_num_exponent_bits = 4
        elif self.config.fp8_format == 'e5m2':
            fp8_num_fraction_bits = 2
            fp8_num_exponent_bits = 5
        else:
            raise ValueError(f'Unknown fp8 format: {self.config.fp8_format}')

        if not torch.is_floating_point(inputs.tensor):
            raise ValueError(f'Expected floating point tensor, got {inputs.tensor.dtype}')

        fp32_num_fraction_bits = 23
        fp32_num_exponent_bits = 8

        fp8_min_exponent = -(1 << (fp8_num_exponent_bits - 1)) + 1
        fp8_max_exponent = (1 << (fp8_num_exponent_bits - 1)) - 1
        fp32_max_exponent = 127

        fp32_fraction_mask = (1 << fp32_num_fraction_bits) - 1
        fp32_exponent_mask = (1 << fp32_num_exponent_bits) - 1
        fp8_fraction_mask = (1 << fp8_num_fraction_bits) - 1

        # Convert to FP32 representation
        values = [ctypes.c_uint32.from_buffer(ctypes.c_float(t)).value for t in inputs.tensor.flatten().to(torch.float16)]

        # Parse FP32 representations
        fp8_fraction = [((v & fp32_fraction_mask) >> (fp32_num_fraction_bits - fp8_num_fraction_bits)) & fp8_fraction_mask for v in values]
        fp32_exponent = [(v >> fp32_num_fraction_bits) & fp32_exponent_mask for v in values]
        fp8_exponent = (torch.clamp(torch.tensor(fp32_exponent) - fp32_max_exponent, min=fp8_min_exponent, max=fp8_max_exponent) + fp8_max_exponent)
        sign_bits = [(v >> 31) & 1 for v in values]

        # Make FP8 values
        # fp8_values = [sign_bits[i] << 7 | fp8_exponent[i] << fp8_num_fraction_bits | fp8_fraction[i] for i in range(len(values))]

        new_fp32_exponent = [v - fp8_max_exponent + fp32_max_exponent for v in fp8_exponent]
        fp32_values = [sign_bits[i] << 31 | new_fp32_exponent[i] << fp32_num_fraction_bits | fp8_fraction[i] << (fp32_num_fraction_bits - fp8_num_fraction_bits) for i in range(len(values))]
        new_tensor = torch.tensor([ctypes.c_float.from_buffer(ctypes.c_uint32(int(v))).value for v in fp32_values], dtype=inputs.tensor.dtype).reshape(inputs.tensor.shape)

        # Overwrite the special values
        new_tensor[torch.isnan(inputs.tensor)] = float('nan')
        new_tensor[torch.isinf(inputs.tensor)] = float('inf')
        new_tensor[torch.isneginf(inputs.tensor)] = float('-inf')
        new_tensor[fp8_exponent.reshape(inputs.tensor.shape) == 0] = 0.0  # Subnormal numbers are considered zero.

        return self.Outputs(tensor=new_tensor)

    def dry_run(self, inputs):
        return self.execute(inputs)



import unittest
import torch
from irisml.tasks.map_int_list import Task


class TestMapIntList(unittest.TestCase):
    def test_simple(self):
        mapping = [torch.tensor([0, 1, 2]), torch.tensor([3, 4, 5])]
        outputs = Task(Task.Config()).execute(Task.Inputs(data=torch.tensor([0, 1, 2, 3, 4, 5]), mapping=mapping))
        self.assertEqual(outputs.data.tolist(), [0, 0, 0, 1, 1, 1])



import dataclasses
import logging
import random
import typing
import PIL.Image
import PIL.ImageColor
import torch.utils.data
import irisml.core

logger = logging.getLogger(__name__)


class Task(irisml.core.TaskBase):
    """Generate a fake image-text classification dataset.

    A generated dataset returns ((image: PIL.Image, text: str), class_id: torch.LongTensor[]) pairs.

    Images are 224x224 pixels, and are filled with a single color. The color is chosen randomly from the list of
    colors in PIL.ImageColor.colormap, and the list is shuffled with the given random seed.

    Config:
        num_images (int): Number of images to generate. Default: 100.
        num_classes (int): Number of classes to generate. Default: 10.
        random_seed (int): Random seed. Default: 0.
    """
    VERSION = '0.1.0'

    @dataclasses.dataclass
    class Config:
        num_images: int = 100
        num_classes: int = 10
        random_seed: int = 0

    @dataclasses.dataclass
    class Outputs:
        dataset: torch.utils.data.Dataset
        num_classes: int
        class_names: typing.List[str]

    class FakeDataset(torch.utils.data.Dataset):
        def __init__(self, num_images, num_classes, random_seed):
            color_names = list(PIL.ImageColor.colormap.keys())
            if num_classes > len(color_names):
                raise ValueError(f"The maximum number of classes is {len(color_names)}")

            rng = random.Random(random_seed)
            self._classes = rng.choices(color_names, k=num_classes)
            self._targets = [rng.randint(0, num_classes - 1) for _ in range(num_images)]

            logger.debug(f"Generated classes: {self._classes} (random_seed={random_seed})")

        @property
        def class_names(self):
            return self._classes

        def __len__(self):
            return len(self._targets)

        def __getitem__(self, index):
            if index >= len(self._targets):
                raise IndexError

            color_name = self._classes[self._targets[index]]
            image = PIL.Image.new('RGB', (224, 224), color=color_name)
            text = f"Color is {color_name}"
            return (image, text), torch.tensor(self._targets[index], dtype=torch.long)

    def execute(self, inputs):
        dataset = Task.FakeDataset(self.config.num_images, self.config.num_classes, self.config.random_seed)
        return self.Outputs(dataset, len(dataset.class_names), dataset.class_names)

    def dry_run(self, inputs):
        return self.execute(inputs)



import dataclasses
import typing
import irisml.core


class Task(irisml.core.TaskBase):
    """Get an item from the given list."""
    VERSION = '0.1.0'

    @dataclasses.dataclass
    class Inputs:
        array: typing.List[typing.Any]

    @dataclasses.dataclass
    class Config:
        index: int

    @dataclasses.dataclass
    class Outputs:
        item: typing.Any = None

    def execute(self, inputs):
        return self.Outputs(inputs.array[self.config.index])

    def dry_run(self, inputs):
        return self.execute(inputs)



import sys
import unittest
import unittest.mock
from irisml.core import TaskBase, TaskDescription
from irisml.tasks.run_profiler import Task


class FakeTask:
    class Task(TaskBase):
        VERSION = '0.1.0'

        def execute(self, inputs):
            return self.Outputs()


class TestRunProfiler(unittest.TestCase):
    def test_run(self):
        tasks = [TaskDescription.from_dict(t) for t in [{'task': 'fake_task'}, {'task': 'fake_task'}]]
        with unittest.mock.patch.dict('sys.modules'):
            sys.modules['irisml.tasks.fake_task'] = FakeTask
            outputs = Task(Task.Config(tasks=tasks)).execute(None)
        self.assertGreater(len(outputs.stats_bytes), 0)



import pathlib
import tempfile
import unittest

from irisml.tasks.load_str_list_jsonl import Task


class TestLoadStrListJsonl(unittest.TestCase):
    def test_simple(self):
        with tempfile.NamedTemporaryFile() as f:
            path = pathlib.Path(f.name)
            path.write_text('{"key": "value1"}\n{"key": "value2"}\n')
            outputs = Task(Task.Config(path=path, key_name="key")).execute(Task.Inputs())
            self.assertEqual(outputs.data, ['value1', 'value2'])



import dataclasses
import sys
import unittest
import unittest.mock
from irisml.core import Context, TaskDescription, TaskBase
from irisml.tasks.search_grid_sequential import Task, SearchSpaceConfig


class FakeTask:
    class Task(TaskBase):
        @dataclasses.dataclass
        class Inputs:
            input1: int
            input2: float

        @dataclasses.dataclass
        class Outputs:
            output: float = 0

        def execute(self, inputs):
            return self.Outputs(inputs.input1 + inputs.input2)


class TestSearchGrid(unittest.TestCase):
    def test_execute(self):
        tasks = [{'task': 'fake_task', 'inputs': {'input1': '$env.VALUE1', 'input2': '$env.VALUE2'}}]
        config = Task.Config(search_space=[SearchSpaceConfig('VALUE1', [1, 3, 5]), SearchSpaceConfig('VALUE2', [0.1, 0.3, 0.5])],
                             tasks=[TaskDescription.from_dict(t) for t in tasks],
                             metrics_output_name='fake_task.output')

        context = Context()
        with unittest.mock.patch.dict('sys.modules'):
            sys.modules['irisml.tasks.fake_task'] = FakeTask
            task = Task(config, context)
            outputs = task.execute(None)

        self.assertEqual(outputs.best_parameters, {'VALUE1': 5, 'VALUE2': 0.5})
        self.assertEqual(context.get_outputs('fake_task').output, 5.5)



import unittest
from irisml.tasks.divide_float import Task


class TestDivideFloat(unittest.TestCase):
    def test_simple(self):
        outputs = Task(Task.Config()).execute(Task.Inputs(3, 2))
        self.assertEqual(outputs.result, 1.5)

    def test_zero_division(self):
        with self.assertRaises(ZeroDivisionError):
            Task(Task.Config()).execute(Task.Inputs(3, 0))



import pathlib
import tempfile
import unittest
import torch
from irisml.tasks.save_jit_model import Task


class TestSaveJitModel(unittest.TestCase):
    def test_simple(self):
        model = torch.nn.Linear(1, 1)
        jit_model = torch.jit.script(model)

        with tempfile.NamedTemporaryFile() as f:
            path = pathlib.Path(f.name)
            Task(Task.Config(path=path)).execute(Task.Inputs(jit_model))

            loaded_model = torch.jit.load(path)
        self.assertEqual(jit_model(torch.tensor([1.0])), loaded_model(torch.tensor([1.0])))



import dataclasses
import logging
import torch.nn
import irisml.core

logger = logging.getLogger(__name__)


class Task(irisml.core.TaskBase):
    """Check Inf/NaN values in model parameters.

    If throw_exception is True, throws an exception if the model parameters contain Inf or NaN.
    """
    VERSION = '0.1.0'

    @dataclasses.dataclass
    class Inputs:
        model: torch.nn.Module

    @dataclasses.dataclass
    class Config:
        throw_exception: bool = False

    @dataclasses.dataclass
    class Outputs:
        has_nan_or_inf: bool

    def execute(self, inputs):
        has_nan_or_infinite = False
        for name, param in inputs.model.named_parameters():
            if not torch.isfinite(param).all():
                logger.info(f"The parameter {name} contains NaN or Inf values")
                has_nan_or_infinite = True

        if has_nan_or_infinite and self.config.throw_exception:
            raise ValueError("The model contains NaN or Inf parameters.")

        return self.Outputs(has_nan_or_infinite)

    def dry_run(self, inputs):
        return self.execute(inputs)



import dataclasses
import pickle
import logging
import irisml.core

logger = logging.getLogger(__name__)


class Task(irisml.core.TaskBase):
    """Pickling an object."""
    VERSION = '0.1.0'

    @dataclasses.dataclass
    class Inputs:
        data: object

    @dataclasses.dataclass
    class Outputs:
        data: bytes

    def execute(self, inputs):
        data = pickle.dumps(inputs.data)
        logger.info("Pickling an object.")
        return self.Outputs(data)

    def dry_run(self, inputs):
        return self.execute(inputs)



import unittest
from irisml.tasks.get_fake_image_classification_dataset import Task


class TestGetFakeImageClassificationDataset(unittest.TestCase):
    def test_simple(self):
        outputs = Task(Task.Config(num_images=100, num_classes=10)).execute(Task.Inputs())
        self.assertEqual(len(outputs.dataset), 100)
        self.assertEqual(len(outputs.class_names), 10)
        class_set = set(int(x[1]) for x in outputs.dataset)
        self.assertEqual(len(class_set), 10)

    def test_empty_classes(self):
        outputs = Task(Task.Config(num_images=100, num_classes=200)).execute(Task.Inputs())
        self.assertEqual(len(outputs.dataset), 100)
        self.assertEqual(len(outputs.class_names), 200)
        class_set = set(int(x[1]) for x in outputs.dataset)
        self.assertEqual(len(class_set), 100)

    def test_random_seed(self):
        outputs1 = Task(Task.Config(num_images=100, num_classes=10, random_seed=1)).execute(Task.Inputs())
        outputs2 = Task(Task.Config(num_images=100, num_classes=10, random_seed=1)).execute(Task.Inputs())
        classes1 = [int(x[1]) for x in outputs1.dataset]
        classes2 = [int(x[1]) for x in outputs2.dataset]
        self.assertEqual(classes1, classes2)

        outputs3 = Task(Task.Config(num_images=100, num_classes=10, random_seed=2)).execute(Task.Inputs())
        classes3 = [int(x[1]) for x in outputs3.dataset]
        self.assertNotEqual(classes1, classes3)



import json
import unittest
import torch
from irisml.tasks.get_int_list_from_json_strings import Task


class TestGetIntListFromJsonStrings(unittest.TestCase):
    def test_simple(self):
        json_strings = [json.dumps({'a': [1, 2, 3], 'b': [3, 4, 5]}), json.dumps({'a': [6, 7, 8], 'b': [9, 10, 11]})]
        outputs = Task(Task.Config(key_name='a')).execute(Task.Inputs(json_strings))
        self.assertTrue(torch.equal(outputs.tensors[0], torch.tensor([1, 2, 3])))
        self.assertTrue(torch.equal(outputs.tensors[1], torch.tensor([6, 7, 8])))

    def test_empty(self):
        json_strings = [json.dumps({'a': [], 'b': [3, 4, 5]}), json.dumps({'b': [9, 10, 11]})]
        outputs = Task(Task.Config(key_name='a')).execute(Task.Inputs(json_strings))
        self.assertTrue(torch.equal(outputs.tensors[0], torch.tensor([])))
        self.assertEqual(outputs.tensors[0].dtype, torch.int64)
        self.assertTrue(torch.equal(outputs.tensors[1], torch.tensor([])))
        self.assertEqual(outputs.tensors[1].dtype, torch.int64)



import dataclasses
import typing
import torch
import irisml.core


class Task(irisml.core.TaskBase):
    """Map a list of integers to a list of integers.

    The mapping is specified as a list of lists of integers. The mapping is
    applied to each element of the input list.
    """
    VERSION = '0.1.0'

    @dataclasses.dataclass
    class Inputs:
        data: torch.Tensor
        mapping: typing.List[torch.Tensor]

    @dataclasses.dataclass
    class Outputs:
        data: torch.Tensor

    def execute(self, inputs):
        if not isinstance(inputs.data, torch.Tensor):
            raise TypeError(f'Expected data to be of type torch.Tensor, got {type(inputs.data)}')
        if not all(isinstance(t, torch.Tensor) for t in inputs.mapping):
            raise TypeError(f'Expected mapping to be a list of torch.Tensor, got {[type(t) for t in inputs.mapping]}')

        if inputs.data.dim() != 1:
            raise ValueError(f'Expected data to be a 1D tensor, got {inputs.data.dim()}D')
        if not all(t.dim() == 1 for t in inputs.mapping):
            raise ValueError(f'Expected mapping to be a list of 1D tensors, got {[t.dim() for t in inputs.mapping]}')

        mapping = {int(v): i for i, t in enumerate(inputs.mapping) for v in t}
        return self.Outputs(data=torch.tensor([mapping[int(v)] for v in inputs.data]))

    def dry_run(self, inputs):
        return self.execute(inputs)



import unittest
import PIL.Image
import torch
from irisml.tasks.get_fake_phrase_grounding_dataset import Task


class TestGetFakePhraseGroundingDataset(unittest.TestCase):
    def test_simple(self):
        outputs = Task(Task.Config(num_images=100, num_max_boxes=10)).execute(None)
        dataset = outputs.dataset
        self.assertEqual(len(dataset), 100)
        self.assertIsInstance(dataset[0][0][0], str)
        self.assertIsInstance(dataset[0][0][1], PIL.Image.Image)
        self.assertIsInstance(dataset[0][1][0], tuple)
        self.assertEqual(len(dataset[0][1][0][0]), 2)
        self.assertIsInstance(dataset[0][1][0][0][0], int)
        self.assertIsInstance(dataset[0][1][0][0][1], int)
        self.assertIsInstance(dataset[0][1][0][1], torch.Tensor)

    def test_multiple_image_sizes(self):
        outputs = Task(Task.Config(image_sizes=[(8, 8), (24, 24), (16, 16), (64, 2)])).execute(None)
        image_sizes_set = set(image.size for (_, image), _ in outputs.dataset)
        self.assertEqual(image_sizes_set, {(8, 8), (24, 24), (16, 16), (64, 2)})

    def test_random_seed(self):
        outputs1 = Task(Task.Config(num_images=100, random_seed=0)).execute(Task.Inputs())
        outputs2 = Task(Task.Config(num_images=100, random_seed=0)).execute(Task.Inputs())

        targets1 = [d[1][0][1].tolist() if d[1] else None for d in outputs1.dataset]
        targets2 = [d[1][0][1].tolist() if d[1] else None for d in outputs2.dataset]
        self.assertEqual(targets1, targets2)

        outputs3 = Task(Task.Config(num_images=100, random_seed=1)).execute(Task.Inputs())
        targets3 = [d[1][0][1].tolist() if d[1] else None for d in outputs3.dataset]
        self.assertNotEqual(targets1, targets3)



import json
import pathlib
import tempfile
import unittest
import torch
from irisml.tasks.save_float_tensor_jsonl import Task


class TestSaveFloatTensorJsonl(unittest.TestCase):
    def test_simple(self):
        input_data = torch.tensor([[1.0, 2.0, 3.0], [4.0, 5.0, 6.0]])
        with tempfile.NamedTemporaryFile() as f:
            temp_file = pathlib.Path(f.name)
            Task(Task.Config(path=temp_file, key_name='float')).execute(Task.Inputs(data=input_data))

            with open(temp_file, 'r') as f:
                lines = [json.loads(line) for line in f]
                self.assertEqual(lines, [{'float': [1.0, 2.0, 3.0]}, {'float': [4.0, 5.0, 6.0]}])



import dataclasses
import logging
import pathlib
import typing
import irisml.core

logger = logging.getLogger(__name__)


class Task(irisml.core.TaskBase):
    """Load lines from a text file.

    Each line is stripped of whitespace and empty lines are removed.

    Config:
        filepath (Path): Path to the text file.
        allow_not_found (bool): If True, allow the file to not exist. If False, raise an error if the file does not exist. Default: False
    """
    VERSION = '0.1.1'
    CACHE_ENABLED = False

    @dataclasses.dataclass
    class Config:
        filepath: pathlib.Path
        allow_not_found: bool = False

    @dataclasses.dataclass
    class Outputs:
        strs: typing.List[str]
        num_strs: int

    def execute(self, inputs):
        if not self.config.filepath.exists():
            if self.config.allow_not_found:
                logger.warning(f"File not found: {self.config.filepath}")
                return self.Outputs(strs=[], num_strs=0)
            else:
                raise FileNotFoundError(f"File not found: {self.config.filepath}")

        logger.info(f"Loading lines from {self.config.filepath}")
        lines = self.config.filepath.read_text().splitlines()
        lines = [line.strip() for line in lines if line.strip()]
        logger.info(f"Loaded {len(lines)} lines from {self.config.filepath}")
        return self.Outputs(strs=lines, num_strs=len(lines))

    def dry_run(self, inputs):
        return self.execute(inputs)



import dataclasses
import logging
import typing
import torch
import irisml.core


logger = logging.getLogger(__name__)


class Task(irisml.core.TaskBase):
    """Calculate cosine similarity between two sets of vectors.

    Inputs:
        tensor0 (torch.Tensor): Shape [N0, M]
        tensor1 (torch.Tensor): Shape [N1, M]

    Config:
        device (str): Device to use. If not specified, it uses cuda if available.

    Outputs:
        cosine_similarity (torch.Tensor): Shape [N0, N1]
    """
    VERSION = '0.1.1'

    @dataclasses.dataclass
    class Inputs:
        tensor0: torch.Tensor
        tensor1: torch.Tensor

    @dataclasses.dataclass
    class Config:
        device: typing.Optional[typing.Literal['cpu', 'cuda']] = None

    @dataclasses.dataclass
    class Outputs:
        cosine_similarity: torch.Tensor

    def execute(self, inputs):
        if inputs.tensor0.shape[1] != inputs.tensor1.shape[1] or len(inputs.tensor0.shape) != 2 or len(inputs.tensor1.shape) != 2:
            raise RuntimeError(f"Input tensors have unexpected shape: tensor0.shape={inputs.tensor0.shape}, tensor1.shape={inputs.tensor1.shape}")

        device = self._get_device()
        tensor0 = inputs.tensor0.to(device)
        tensor1 = inputs.tensor1.to(device)

        eps = 1e-8
        norm0 = tensor0.norm(dim=1)[:, None]
        norm1 = tensor1.norm(dim=1)[:, None]
        normalized_tensor0 = tensor0 / torch.clamp(norm0, min=eps)
        normalized_tensor1 = tensor1 / torch.clamp(norm1, min=eps)
        cosine_similarity = torch.mm(normalized_tensor0, normalized_tensor1.transpose(0, 1)).to('cpu')
        assert cosine_similarity.shape == (inputs.tensor0.shape[0], inputs.tensor1.shape[0])
        return self.Outputs(cosine_similarity)

    def dry_run(self, inputs):
        fake_result = torch.zeros((inputs.tensor0.shape[0], inputs.tensor1.shape[0]))
        return self.Outputs(fake_result)

    def _get_device(self) -> torch.device:
        """Get a torch device based on the configuration. If not specified explicitly, it uses cuda if available."""
        if self.config.device:
            device_name = self.config.device
        else:
            device_name = 'cuda' if torch.cuda.is_available() else 'cpu'
            logger.info(f"Device is selected automatically: {device_name}. To specify the device manually, please set Config.device.")

        return torch.device(device_name)



import unittest
from irisml.tasks.join_two_strings import Task


class TestJoinTwoStrings(unittest.TestCase):
    def test_simple(self):
        config = Task.Config(delimiter=' and ')
        outputs = Task(config).execute(Task.Inputs('1st str', '2nd str'))
        self.assertEqual(outputs.string, '1st str and 2nd str')

    def test_no_string2(self):
        first_string = '1st str'
        outputs = Task(Task.Config()).execute(Task.Inputs(first_string))
        self.assertEqual(outputs.string, first_string)

    def test_two_empty_strings(self):
        with self.assertRaises(ValueError):
            Task(Task.Config()).execute(Task.Inputs('', ''))



import dataclasses
import pickle
import pathlib
import typing
import torch
import irisml.core


class Task(irisml.core.TaskBase):
    """Load a list of tensors from file."""
    VERSION = '0.1.0'

    @dataclasses.dataclass
    class Inputs:
        path: pathlib.Path

    @dataclasses.dataclass
    class Outputs:
        tensor_list: typing.List[torch.Tensor]

    def execute(self, inputs):
        data = pickle.loads(inputs.path.read_bytes())
        if not isinstance(data, list):
            raise TypeError(f"Loaded data is type {type(data)}, expected list")
        for i, t in enumerate(data):
            if not torch.is_tensor(t):
                raise TypeError(f"item {i} in the list is type {type(t)}, expected tensor")
        return self.Outputs(data)

    def dry_run(self, inputs):
        return self.execute(inputs)



import dataclasses
import logging
import pprint
from typing import Dict, List, Optional
import torch
import irisml.core

logger = logging.getLogger(__name__)


class Task(irisml.core.TaskBase):
    """Print or Pretty Print the input object."""
    VERSION = '0.1.1'
    CACHE_ENABLED = False

    @dataclasses.dataclass
    class Inputs:
        data_dict_str_float: Optional[Dict[str, float]] = None
        data_float: Optional[float] = None
        data_int: Optional[int] = None
        data_tensor: Optional[torch.Tensor] = None
        data_list_str: Optional[List[str]] = None

    @dataclasses.dataclass
    class Config:
        label: Optional[str] = None
        pretty: bool = False

    def execute(self, inputs):
        data_all = [getattr(inputs, f.name) for f in dataclasses.fields(inputs)]
        data_provided = [x for x in data_all if x is not None]
        if len(data_provided) != 1:
            raise ValueError("Zero or multiple inputs are provided.")

        if self.config.label:
            logger.info(f"LABEL: {self.config.label}")

        data = data_provided[0]
        if isinstance(data, torch.Tensor):
            logger.info(f"Tensor shape={data.shape}, dtype={data.dtype}")

        if self.config.pretty:
            logger.info(pprint.pformat(data))
        else:
            logger.info(str(data))

        return self.Outputs()

    def dry_run(self, inputs):
        return self.execute(inputs)



import dataclasses
import logging
import random
import typing
import PIL.Image
import PIL.ImageDraw
import torch
import irisml.core

logger = logging.getLogger(__name__)


class Task(irisml.core.TaskBase):
    """Generate a fake image classification dataset with multiple labels per image.

    The dataset consists of 224x224 RGB images with random color boxes drawn on them.

    Config:
        num_images (int): Number of images to generate. Default: 100
        num_classes (int): Number of classes to generate. Default: 10
        random_seed (int): Random seed. Default: 0
    """
    VERSION = '0.1.0'

    @dataclasses.dataclass
    class Config:
        num_images: int = 100
        num_classes: int = 10
        random_seed: int = 0

    @dataclasses.dataclass
    class Outputs:
        dataset: torch.utils.data.Dataset
        num_classes: int
        class_names: typing.List[str]

    class FakeDataset(torch.utils.data.Dataset):
        def __init__(self, image_labels: typing.List[typing.List[int]], colors: typing.List[typing.Tuple[int, int, int]], random_seed: int):
            self._image_labels = image_labels
            self._colors = colors
            self._random_seed = random_seed
            rng = random.Random(self._random_seed)
            box_center_xy = [(rng.randint(0, 224), rng.randint(0, 224)) for _ in range(len(colors))]
            self._boxes = [(max(0, x - 16), max(0, y - 16), min(224, x + 16), min(224, y + 16)) for x, y in box_center_xy]

        def __len__(self):
            return len(self._image_labels)

        def __getitem__(self, index):
            targets = self._image_labels[index]
            return self._generate_image(targets), torch.zeros(len(self._colors), dtype=torch.int).scatter_(0, torch.tensor(targets), 1)

        def _generate_image(self, targets):
            image = PIL.Image.new('RGB', (224, 224), color=(255, 255, 255))
            draw = PIL.ImageDraw.Draw(image)
            for i, target in enumerate(targets):
                draw.rectangle(self._boxes[i], fill=self._colors[target])
            return image

    def execute(self, inputs):
        if not self.config.num_images or not self.config.num_classes:
            raise ValueError(f"Invalid settings: num_images={self.config.num_images}, num_classes={self.config.num_classes}")

        if self.config.num_images < self.config.num_classes:
            logger.info(f"Some classes will be empty: num_images={self.config.num_images}, num_classes={self.config.num_classes}")

        rng = random.Random(self.config.random_seed)
        image_labels = [random.sample(range(self.config.num_classes), rng.randint(0, self.config.num_classes)) for _ in range(self.config.num_images)]
        colors = [(rng.randint(0, 255), rng.randint(0, 255), rng.randint(0, 255)) for _ in range(self.config.num_classes)]

        return self.Outputs(
            dataset=self.FakeDataset(image_labels, colors, self.config.random_seed),
            num_classes=self.config.num_classes,
            class_names=[f'class_{i}' for i in range(self.config.num_classes)],
        )

    def dry_run(self, inputs):
        return self.execute(inputs)
